#    Copyright 2018 Rackspace, US Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from cryptography import fernet
from jsonschema import exceptions as js_exceptions
from jsonschema import validate

from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging as messaging
from stevedore import driver as stevedore_driver

from octavia_lib.api.drivers import data_models as driver_dm
from octavia_lib.api.drivers import exceptions
from octavia_lib.api.drivers import provider_base as driver_base
from octavia_lib.common import constants as lib_consts

from octavia.api.drivers.amphora_driver import availability_zone_schema
from octavia.api.drivers.amphora_driver import flavor_schema
from octavia.api.drivers import utils as driver_utils
from octavia.common import constants as consts
from octavia.common import data_models
from octavia.common import rpc
from octavia.common import utils
from octavia.db import api as db_apis
from octavia.db import repositories
from octavia.network import base as network_base

CONF = cfg.CONF
CONF.import_group('oslo_messaging', 'octavia.common.config')
LOG = logging.getLogger(__name__)
AMPHORA_SUPPORTED_LB_ALGORITHMS = [
    consts.LB_ALGORITHM_ROUND_ROBIN,
    consts.LB_ALGORITHM_SOURCE_IP,
    consts.LB_ALGORITHM_LEAST_CONNECTIONS]

AMPHORA_SUPPORTED_PROTOCOLS = [
    lib_consts.PROTOCOL_TCP,
    lib_consts.PROTOCOL_HTTP,
    lib_consts.PROTOCOL_HTTPS,
    lib_consts.PROTOCOL_TERMINATED_HTTPS,
    lib_consts.PROTOCOL_PROXY,
    lib_consts.PROTOCOL_PROXYV2,
    lib_consts.PROTOCOL_UDP,
    lib_consts.PROTOCOL_SCTP,
    lib_consts.PROTOCOL_PROMETHEUS,
]

VALID_L7POLICY_LISTENER_PROTOCOLS = [
    lib_consts.PROTOCOL_HTTP,
    lib_consts.PROTOCOL_TERMINATED_HTTPS
]


class AmphoraProviderDriver(driver_base.ProviderDriver):
    def __init__(self):
        super().__init__()
        self.target = messaging.Target(
            namespace=consts.RPC_NAMESPACE_CONTROLLER_AGENT,
            topic=consts.TOPIC_AMPHORA_V2, version="2.0", fanout=False)
        self.client = rpc.get_client(self.target)
        self.repositories = repositories.Repositories()
        key = utils.get_compatible_server_certs_key_passphrase()
        self.fernet = fernet.Fernet(key)

    def _validate_pool_algorithm(self, pool):
        if pool.lb_algorithm not in AMPHORA_SUPPORTED_LB_ALGORITHMS:
            msg = (f'Amphora provider does not support {pool.lb_algorithm} '
                   f'algorithm.')
            raise exceptions.UnsupportedOptionError(
                user_fault_string=msg,
                operator_fault_string=msg)

    def _validate_listener_protocol(self, listener):
        if listener.protocol not in AMPHORA_SUPPORTED_PROTOCOLS:
            msg = ('Amphora provider does not support %s protocol. '
                   'Supported: %s'
                   % (listener.protocol,
                      ", ".join(AMPHORA_SUPPORTED_PROTOCOLS)))
            raise exceptions.UnsupportedOptionError(
                user_fault_string=msg,
                operator_fault_string=msg)

    def _validate_alpn_protocols(self, obj):
        if not obj.alpn_protocols:
            return
        supported = consts.AMPHORA_SUPPORTED_ALPN_PROTOCOLS
        not_supported = set(obj.alpn_protocols) - set(supported)
        if not_supported:
            msg = ('Amphora provider does not support %s ALPN protocol(s). '
                   'Supported: %s'
                   % (", ".join(not_supported), ", ".join(supported)))
            raise exceptions.UnsupportedOptionError(
                user_fault_string=msg,
                operator_fault_string=msg)

    # Load Balancer
    def create_vip_port(self, loadbalancer_id, project_id, vip_dictionary,
                        additional_vip_dicts):
        vip_obj = driver_utils.provider_vip_dict_to_vip_obj(vip_dictionary)
        add_vip_objs = [
            driver_utils.provider_additional_vip_dict_to_additional_vip_obj(
                add_vip)
            for add_vip in additional_vip_dicts]
        lb_obj = data_models.LoadBalancer(id=loadbalancer_id,
                                          project_id=project_id, vip=vip_obj,
                                          additional_vips=add_vip_objs)

        network_driver = utils.get_network_driver()
        vip_network = network_driver.get_network(
            vip_dictionary[lib_consts.VIP_NETWORK_ID])
        if not vip_network.port_security_enabled:
            message = "Port security must be enabled on the VIP network."
            raise exceptions.DriverError(user_fault_string=message,
                                         operator_fault_string=message)

        try:
            vip, add_vips = network_driver.allocate_vip(lb_obj)
        except network_base.AllocateVIPException as e:
            message = str(e)
            if getattr(e, 'orig_msg', None) is not None:
                message = e.orig_msg
            raise exceptions.DriverError(user_fault_string=message,
                                         operator_fault_string=message)

        LOG.info('Amphora provider created VIP port %s for load balancer %s.',
                 vip.port_id, loadbalancer_id)
        vip_return_dict = driver_utils.vip_dict_to_provider_dict(vip.to_dict())
        add_return_dicts = [driver_utils.additional_vip_dict_to_provider_dict(
            add_vip.to_dict()) for add_vip in add_vips]
        return vip_return_dict, add_return_dicts

    # TODO(johnsom) convert this to octavia_lib constant flavor
    # once octavia is transitioned to use octavia_lib
    def loadbalancer_create(self, loadbalancer):
        if loadbalancer.flavor == driver_dm.Unset:
            loadbalancer.flavor = None
        if loadbalancer.availability_zone == driver_dm.Unset:
            loadbalancer.availability_zone = None
        payload = {consts.LOADBALANCER: loadbalancer.to_dict(),
                   consts.FLAVOR: loadbalancer.flavor,
                   consts.AVAILABILITY_ZONE: loadbalancer.availability_zone}
        self.client.cast({}, 'create_load_balancer', **payload)

    def loadbalancer_delete(self, loadbalancer, cascade=False):
        payload = {consts.LOADBALANCER: loadbalancer.to_dict(),
                   'cascade': cascade}
        self.client.cast({}, 'delete_load_balancer', **payload)

    def loadbalancer_failover(self, loadbalancer_id):
        payload = {consts.LOAD_BALANCER_ID: loadbalancer_id}
        self.client.cast({}, 'failover_load_balancer', **payload)

    def loadbalancer_update(self, original_load_balancer, new_loadbalancer):
        # Adapt the provider data model to the queue schema
        lb_dict = new_loadbalancer.to_dict()
        if 'admin_state_up' in lb_dict:
            lb_dict['enabled'] = lb_dict.pop('admin_state_up')
        # Put the qos_policy_id back under the vip element the controller
        # expects
        vip_qos_policy_id = lb_dict.pop('vip_qos_policy_id', None)
        lb_dict.pop(consts.LOADBALANCER_ID)
        if vip_qos_policy_id:
            vip_dict = {"qos_policy_id": vip_qos_policy_id}
            lb_dict["vip"] = vip_dict

        payload = {consts.ORIGINAL_LOADBALANCER:
                   original_load_balancer.to_dict(),
                   consts.LOAD_BALANCER_UPDATES: lb_dict}
        self.client.cast({}, 'update_load_balancer', **payload)

    def _encrypt_tls_container_data(self, tls_container_data):
        for key, val in tls_container_data.items():
            if isinstance(val, bytes):
                tls_container_data[key] = self.fernet.encrypt(val)
            elif isinstance(val, list):
                encrypt_vals = []
                for i in val:
                    if isinstance(i, bytes):
                        encrypt_vals.append(self.fernet.encrypt(i))
                    else:
                        encrypt_vals.append(i)
                tls_container_data[key] = encrypt_vals

    def _encrypt_listener_dict(self, listener_dict):
        # We need to encrypt the user cert/key data for sending it
        # over messaging.
        if listener_dict.get(consts.DEFAULT_TLS_CONTAINER_DATA, False):
            container_data = listener_dict[consts.DEFAULT_TLS_CONTAINER_DATA]
            self._encrypt_tls_container_data(container_data)
        if listener_dict.get(consts.SNI_CONTAINER_DATA, False):
            sni_list = []
            for sni_data in listener_dict[consts.SNI_CONTAINER_DATA]:
                self._encrypt_tls_container_data(sni_data)
                sni_list.append(sni_data)
            if sni_list:
                listener_dict[consts.SNI_CONTAINER_DATA] = sni_list

    # Listener
    def listener_create(self, listener):
        self._validate_listener_protocol(listener)
        self._validate_alpn_protocols(listener)
        payload = {consts.LISTENER: listener.to_dict()}
        self._encrypt_listener_dict(payload[consts.LISTENER])

        self.client.cast({}, 'create_listener', **payload)

    def listener_delete(self, listener):
        payload = {consts.LISTENER: listener.to_dict()}
        self.client.cast({}, 'delete_listener', **payload)

    def listener_update(self, old_listener, new_listener):
        self._validate_alpn_protocols(new_listener)
        original_listener = old_listener.to_dict()
        listener_updates = new_listener.to_dict()

        self._encrypt_listener_dict(original_listener)
        self._encrypt_listener_dict(listener_updates)

        payload = {consts.ORIGINAL_LISTENER: original_listener,
                   consts.LISTENER_UPDATES: listener_updates}
        self.client.cast({}, 'update_listener', **payload)

    # Pool
    def _pool_convert_to_dict(self, pool):
        pool_dict = pool.to_dict(recurse=True)
        if 'admin_state_up' in pool_dict:
            pool_dict['enabled'] = pool_dict.pop('admin_state_up')
        if 'tls_container_ref' in pool_dict:
            pool_dict['tls_certificate_id'] = pool_dict.pop(
                'tls_container_ref')
        pool_dict.pop('tls_container_data', None)
        if 'ca_tls_container_ref' in pool_dict:
            pool_dict['ca_tls_certificate_id'] = pool_dict.pop(
                'ca_tls_container_ref')
        pool_dict.pop('ca_tls_container_data', None)
        if 'crl_container_ref' in pool_dict:
            pool_dict['crl_container_id'] = pool_dict.pop('crl_container_ref')
        pool_dict.pop('crl_container_data', None)
        return pool_dict

    def pool_create(self, pool):
        self._validate_pool_algorithm(pool)
        self._validate_alpn_protocols(pool)
        payload = {consts.POOL: self._pool_convert_to_dict(pool)}
        self.client.cast({}, 'create_pool', **payload)

    def pool_delete(self, pool):
        payload = {consts.POOL: pool.to_dict(recurse=True)}
        self.client.cast({}, 'delete_pool', **payload)

    def pool_update(self, old_pool, new_pool):
        self._validate_alpn_protocols(new_pool)
        if new_pool.lb_algorithm:
            self._validate_pool_algorithm(new_pool)
        pool_dict = self._pool_convert_to_dict(new_pool)
        pool_dict.pop('pool_id')
        payload = {consts.ORIGINAL_POOL: old_pool.to_dict(),
                   consts.POOL_UPDATES: pool_dict}
        self.client.cast({}, 'update_pool', **payload)

    # Member
    def member_create(self, member):
        pool_id = member.pool_id
        session = db_apis.get_session()
        with session.begin():
            db_pool = self.repositories.pool.get(session,
                                                 id=pool_id)
        self._validate_members(db_pool, [member])

        payload = {consts.MEMBER: member.to_dict()}
        self.client.cast({}, 'create_member', **payload)

    def member_delete(self, member):
        payload = {consts.MEMBER: member.to_dict()}
        self.client.cast({}, 'delete_member', **payload)

    def member_update(self, old_member, new_member):
        original_member = old_member.to_dict()
        member_updates = new_member.to_dict()
        if 'admin_state_up' in member_updates:
            member_updates['enabled'] = member_updates.pop('admin_state_up')
        member_updates.pop(consts.MEMBER_ID)
        payload = {consts.ORIGINAL_MEMBER: original_member,
                   consts.MEMBER_UPDATES: member_updates}
        self.client.cast({}, 'update_member', **payload)

    def member_batch_update(self, pool_id, members):
        # The DB should not have updated yet, so we can still use the pool
        session = db_apis.get_session()
        with session.begin():
            db_pool = self.repositories.pool.get(session, id=pool_id)

        self._validate_members(db_pool, members)

        old_members = db_pool.members

        old_member_ids = [m.id for m in old_members]
        # The driver will always pass objects with IDs.
        new_member_ids = [m.member_id for m in members]

        # Find members that are brand new or updated
        new_members = []
        updated_members = []
        for m in members:
            if m.member_id not in old_member_ids:
                new_members.append(m)
            else:
                member_dict = m.to_dict(render_unsets=False)
                member_dict['id'] = member_dict.pop('member_id')
                if 'address' in member_dict:
                    member_dict['ip_address'] = member_dict.pop('address')
                if 'admin_state_up' in member_dict:
                    member_dict['enabled'] = member_dict.pop('admin_state_up')
                updated_members.append(member_dict)

        # Find members that are deleted
        deleted_members = []
        for m in old_members:
            if m.id not in new_member_ids:
                deleted_members.append(m)

        payload = {'old_members': [m.to_dict() for m in deleted_members],
                   'new_members': [m.to_dict() for m in new_members],
                   'updated_members': updated_members}
        self.client.cast({}, 'batch_update_members', **payload)

    def _validate_members(self, db_pool, members):
        if db_pool.protocol in consts.LVS_PROTOCOLS:
            # For SCTP/UDP LBs:
            # Allow ipv4 member if there's at least one ipv4 VIP
            # Allow ipv6 member if there's at least one ipv6 VIP
            for member in members:
                member_is_ipv6 = utils.is_ipv6(member.address)

                for listener in db_pool.listeners:
                    lb = listener.load_balancer
                    vips = [lb.vip]
                    vips.extend(lb.additional_vips)
                    lb_has_ipv4 = any(utils.is_ipv4(vip.ip_address)
                                      for vip in vips)
                    lb_has_ipv6 = any(utils.is_ipv6(vip.ip_address)
                                      for vip in vips)

                    if ((member_is_ipv6 and not lb_has_ipv6) or
                            (not member_is_ipv6 and not lb_has_ipv4)):
                        msg = ("This provider doesn't support mixing IPv4 and "
                               "IPv6 addresses for its VIP and members in {} "
                               "load balancers.".format(db_pool.protocol))
                        raise exceptions.UnsupportedOptionError(
                            user_fault_string=msg,
                            operator_fault_string=msg)

    # Health Monitor
    def health_monitor_create(self, healthmonitor):
        payload = {consts.HEALTH_MONITOR: healthmonitor.to_dict()}
        self.client.cast({}, 'create_health_monitor', **payload)

    def health_monitor_delete(self, healthmonitor):
        payload = {consts.HEALTH_MONITOR: healthmonitor.to_dict()}
        self.client.cast({}, 'delete_health_monitor', **payload)

    def health_monitor_update(self, old_healthmonitor, new_healthmonitor):
        healthmon_dict = new_healthmonitor.to_dict()
        if 'admin_state_up' in healthmon_dict:
            healthmon_dict['enabled'] = healthmon_dict.pop('admin_state_up')
        if 'max_retries_down' in healthmon_dict:
            healthmon_dict['fall_threshold'] = healthmon_dict.pop(
                'max_retries_down')
        if 'max_retries' in healthmon_dict:
            healthmon_dict['rise_threshold'] = healthmon_dict.pop(
                'max_retries')
        healthmon_dict.pop('healthmonitor_id')

        payload = {consts.ORIGINAL_HEALTH_MONITOR: old_healthmonitor.to_dict(),
                   consts.HEALTH_MONITOR_UPDATES: healthmon_dict}
        self.client.cast({}, 'update_health_monitor', **payload)

    # L7 Policy
    def l7policy_create(self, l7policy):
        session = db_apis.get_session()
        with session.begin():
            db_listener = self.repositories.listener.get(
                session, id=l7policy.listener_id)
        if db_listener.protocol not in VALID_L7POLICY_LISTENER_PROTOCOLS:
            msg = ('%s protocol listeners do not support L7 policies' % (
                db_listener.protocol))
            raise exceptions.UnsupportedOptionError(
                user_fault_string=msg,
                operator_fault_string=msg)
        payload = {consts.L7POLICY: l7policy.to_dict()}
        self.client.cast({}, 'create_l7policy', **payload)

    def l7policy_delete(self, l7policy):
        payload = {consts.L7POLICY: l7policy.to_dict()}
        self.client.cast({}, 'delete_l7policy', **payload)

    def l7policy_update(self, old_l7policy, new_l7policy):
        l7policy_dict = new_l7policy.to_dict()
        if 'admin_state_up' in l7policy_dict:
            l7policy_dict['enabled'] = l7policy_dict.pop(consts.ADMIN_STATE_UP)
        l7policy_dict.pop(consts.L7POLICY_ID)

        payload = {consts.ORIGINAL_L7POLICY: old_l7policy.to_dict(),
                   consts.L7POLICY_UPDATES: l7policy_dict}
        self.client.cast({}, 'update_l7policy', **payload)

    # L7 Rule
    def l7rule_create(self, l7rule):
        payload = {consts.L7RULE: l7rule.to_dict()}
        self.client.cast({}, 'create_l7rule', **payload)

    def l7rule_delete(self, l7rule):
        payload = {consts.L7RULE: l7rule.to_dict()}
        self.client.cast({}, 'delete_l7rule', **payload)

    def l7rule_update(self, old_l7rule, new_l7rule):
        l7rule_dict = new_l7rule.to_dict()
        if consts.ADMIN_STATE_UP in l7rule_dict:
            l7rule_dict['enabled'] = l7rule_dict.pop(consts.ADMIN_STATE_UP)
        l7rule_dict.pop(consts.L7RULE_ID)

        payload = {consts.ORIGINAL_L7RULE: old_l7rule.to_dict(),
                   consts.L7RULE_UPDATES: l7rule_dict}
        self.client.cast({}, 'update_l7rule', **payload)

    # Flavor
    def get_supported_flavor_metadata(self):
        """Returns the valid flavor metadata keys and descriptions.

        This extracts the valid flavor metadata keys and descriptions
        from the JSON validation schema and returns it as a dictionary.

        :return: Dictionary of flavor metadata keys and descriptions.
        :raises DriverError: An unexpected error occurred.
        """
        try:
            props = flavor_schema.SUPPORTED_FLAVOR_SCHEMA['properties']
            return {k: v.get('description', '') for k, v in props.items()}
        except Exception as e:
            raise exceptions.DriverError(
                user_fault_string='Failed to get the supported flavor '
                                  'metadata due to: {}'.format(str(e)),
                operator_fault_string='Failed to get the supported flavor '
                                      'metadata due to: {}'.format(str(e)))

    def validate_flavor(self, flavor_dict):
        """Validates flavor profile data.

        This will validate a flavor profile dataset against the flavor
        settings the amphora driver supports.

        :param flavor_dict: The flavor dictionary to validate.
        :type flavor: dict
        :return: None
        :raises DriverError: An unexpected error occurred.
        :raises UnsupportedOptionError: If the driver does not support
          one of the flavor settings.
        """
        try:
            validate(flavor_dict, flavor_schema.SUPPORTED_FLAVOR_SCHEMA)
        except js_exceptions.ValidationError as e:
            error_object = ''
            if e.relative_path:
                error_object = f'{e.relative_path[0]} '
            raise exceptions.UnsupportedOptionError(
                user_fault_string=f'{error_object}{e.message}',
                operator_fault_string=str(e))
        except Exception as e:
            raise exceptions.DriverError(
                user_fault_string='Failed to validate the flavor metadata '
                                  'due to: {}'.format(str(e)),
                operator_fault_string='Failed to validate the flavor metadata '
                                      'due to: {}'.format(str(e)))
        compute_flavor = flavor_dict.get(consts.COMPUTE_FLAVOR, None)
        if compute_flavor:
            compute_driver = stevedore_driver.DriverManager(
                namespace='octavia.compute.drivers',
                name=CONF.controller_worker.compute_driver,
                invoke_on_load=True
            ).driver

            # TODO(johnsom) Fix this to raise a NotFound error
            # when the octavia-lib supports it.
            compute_driver.validate_flavor(compute_flavor)

        amp_image_tag = flavor_dict.get(consts.AMP_IMAGE_TAG, None)
        if amp_image_tag:
            image_driver = stevedore_driver.DriverManager(
                namespace='octavia.image.drivers',
                name=CONF.controller_worker.image_driver,
                invoke_on_load=True
            ).driver

            try:
                image_driver.get_image_id_by_tag(
                    amp_image_tag, CONF.controller_worker.amp_image_owner_id)
            except Exception as e:
                raise exceptions.NotFound(
                    user_fault_string='Failed to find an image with tag {} '
                                      'due to: {}'.format(
                                          amp_image_tag, str(e)),
                    operator_fault_string='Failed to find an image with tag '
                                          '{} due to: {}'.format(
                                              amp_image_tag, str(e)))

    # Availability Zone
    def get_supported_availability_zone_metadata(self):
        """Returns the valid availability zone metadata keys and descriptions.

        This extracts the valid availability zone metadata keys and
        descriptions from the JSON validation schema and returns it as a
        dictionary.

        :return: Dictionary of availability zone metadata keys and descriptions
        :raises DriverError: An unexpected error occurred.
        """
        try:
            props = (
                availability_zone_schema.SUPPORTED_AVAILABILITY_ZONE_SCHEMA[
                    'properties'])
            return {k: v.get('description', '') for k, v in props.items()}
        except Exception as e:
            raise exceptions.DriverError(
                user_fault_string='Failed to get the supported availability '
                                  'zone metadata due to: {}'.format(str(e)),
                operator_fault_string='Failed to get the supported '
                                      'availability zone metadata due to: '
                                      '{}'.format(str(e)))

    def validate_availability_zone(self, availability_zone_dict):
        """Validates availability zone profile data.

        This will validate an availability zone profile dataset against the
        availability zone settings the amphora driver supports.

        :param availability_zone_dict: The availability zone dict to validate.
        :type availability_zone_dict: dict
        :return: None
        :raises DriverError: An unexpected error occurred.
        :raises UnsupportedOptionError: If the driver does not support
          one of the availability zone settings.
        """
        try:
            validate(
                availability_zone_dict,
                availability_zone_schema.SUPPORTED_AVAILABILITY_ZONE_SCHEMA)
        except js_exceptions.ValidationError as e:
            error_object = ''
            if e.relative_path:
                error_object = f'{e.relative_path[0]} '
            raise exceptions.UnsupportedOptionError(
                user_fault_string=f'{error_object}{e.message}',
                operator_fault_string=str(e))
        except Exception as e:
            raise exceptions.DriverError(
                user_fault_string='Failed to validate the availability zone '
                                  'metadata due to: {}'.format(str(e)),
                operator_fault_string='Failed to validate the availability '
                                      'zone metadata due to: {}'.format(str(e))
            )
        compute_zone = availability_zone_dict.get(consts.COMPUTE_ZONE, None)
        if compute_zone:
            compute_driver = stevedore_driver.DriverManager(
                namespace='octavia.compute.drivers',
                name=CONF.controller_worker.compute_driver,
                invoke_on_load=True
            ).driver

            # TODO(johnsom) Fix this to raise a NotFound error
            # when the octavia-lib supports it.
            compute_driver.validate_availability_zone(compute_zone)

        check_nets = availability_zone_dict.get(
            consts.VALID_VIP_NETWORKS, [])
        management_net = availability_zone_dict.get(
            consts.MANAGEMENT_NETWORK, None)
        if management_net:
            check_nets.append(management_net)
        for check_net in check_nets:
            network_driver = utils.get_network_driver()

            # TODO(johnsom) Fix this to raise a NotFound error
            # when the octavia-lib supports it.
            network_driver.get_network(check_net)



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from collections import namedtuple

from oslo_log import log as logging
from oslo_utils import uuidutils
from sqlalchemy import Column
from sqlalchemy import create_engine
from sqlalchemy import MetaData
from sqlalchemy import String
from sqlalchemy import Table
from sqlalchemy import update

from octavia.common import constants
from octavia.common import data_models
from octavia.compute import compute_base as driver_base
from octavia.network import data_models as network_models

LOG = logging.getLogger(__name__)


NoopServerGroup = namedtuple('ServerGroup', ['id'])


class NoopManager:
    def __init__(self):
        super().__init__()
        self.computeconfig = {}

        # Get a DB engine for the network no-op DB
        # Required to update the ports when a port is attached to a compute
        self.engine = create_engine('sqlite:////tmp/octavia-network-noop.db',
                                    isolation_level='SERIALIZABLE')
        metadata_obj = MetaData()

        self.interfaces_table = Table(
            'interfaces',
            metadata_obj,
            Column('port_id', String(36)),
            Column('network_id', String(36)),
            Column('compute_id', String(36)),
            Column('vnic_type', String(6)))

    def build(self, name="amphora_name", amphora_flavor=None,
              image_tag=None, image_owner=None, key_name=None, sec_groups=None,
              network_ids=None, config_drive_files=None, user_data=None,
              port_ids=None, server_group_id=None, availability_zone=None):
        LOG.debug("Compute %s no-op, build name %s, amphora_flavor %s, "
                  "image_tag %s, image_owner %s, key_name %s, sec_groups %s, "
                  "network_ids %s, config_drive_files %s, user_data %s, "
                  "port_ids %s, server_group_id %s, availability_zone %s",
                  self.__class__.__name__,
                  name, amphora_flavor, image_tag, image_owner,
                  key_name, sec_groups, network_ids, config_drive_files,
                  user_data, port_ids, server_group_id, availability_zone)
        self.computeconfig[(name, amphora_flavor, image_tag,
                            image_owner, key_name, user_data,
                            server_group_id)] = (
            name, amphora_flavor,
            image_tag, image_owner, key_name, sec_groups,
            network_ids, config_drive_files, user_data, port_ids,
            server_group_id, 'build')
        compute_id = uuidutils.generate_uuid()
        return compute_id

    def delete(self, compute_id):
        LOG.debug("Compute %s no-op, compute_id %s",
                  self.__class__.__name__, compute_id)
        self.computeconfig[compute_id] = (compute_id, 'delete')

    def status(self, compute_id):
        LOG.debug("Compute %s no-op, compute_id %s",
                  self.__class__.__name__, compute_id)
        self.computeconfig[compute_id] = (compute_id, 'status')
        return constants.UP

    def get_amphora(self, compute_id, management_network_id=None):
        LOG.debug("Compute %s no-op, compute_id %s, management_network_id %s",
                  self.__class__.__name__, compute_id, management_network_id)
        self.computeconfig[(compute_id, management_network_id)] = (
            compute_id, management_network_id, 'get_amphora')
        return data_models.Amphora(
            compute_id=compute_id,
            status=constants.ACTIVE,
            lb_network_ip='192.0.2.1'
        ), None

    def create_server_group(self, name, policy):
        LOG.debug("Create Server Group %s no-op, name %s, policy %s ",
                  self.__class__.__name__, name, policy)
        self.computeconfig[(name, policy)] = (name, policy, 'create')
        return NoopServerGroup(id=uuidutils.generate_uuid())

    def delete_server_group(self, server_group_id):
        LOG.debug("Delete Server Group %s no-op, id %s ",
                  self.__class__.__name__, server_group_id)
        self.computeconfig[server_group_id] = (server_group_id, 'delete')

    def attach_network_or_port(self, compute_id, network_id=None,
                               ip_address=None, port_id=None):
        LOG.debug("Compute %s no-op, attach_network_or_port compute_id %s,"
                  "network_id %s, ip_address %s, port_id %s",
                  self.__class__.__name__, compute_id,
                  network_id, ip_address, port_id)
        self.computeconfig[(compute_id, network_id, ip_address, port_id)] = (
            compute_id, network_id, ip_address, port_id,
            'attach_network_or_port')

        # Update the port in the network no-op DB
        with self.engine.connect() as connection:
            connection.execute(update(self.interfaces_table).where(
                self.interfaces_table.c.port_id == port_id).values(
                    compute_id=compute_id))
            connection.commit()

        return network_models.Interface(
            id=uuidutils.generate_uuid(),
            compute_id=compute_id,
            network_id=network_id,
            fixed_ips=[],
            port_id=uuidutils.generate_uuid(),
            vnic_type=constants.VNIC_TYPE_NORMAL
        )

    def detach_port(self, compute_id, port_id):
        LOG.debug("Compute %s no-op, detach_network compute_id %s, "
                  "port_id %s",
                  self.__class__.__name__, compute_id, port_id)
        self.computeconfig[(compute_id, port_id)] = (
            compute_id, port_id, 'detach_port')

    def validate_flavor(self, flavor_id):
        LOG.debug("Compute %s no-op, validate_flavor flavor_id %s",
                  self.__class__.__name__, flavor_id)
        self.computeconfig[flavor_id] = (flavor_id, 'validate_flavor')

    def validate_availability_zone(self, availability_zone):
        LOG.debug("Compute %s no-op, validate_availability_zone name %s",
                  self.__class__.__name__, availability_zone)
        self.computeconfig[availability_zone] = (
            availability_zone, 'validate_availability_zone')


class NoopComputeDriver(driver_base.ComputeBase):
    def __init__(self):
        super().__init__()
        self.driver = NoopManager()

    def build(self, name="amphora_name", amphora_flavor=None,
              image_tag=None, image_owner=None, key_name=None, sec_groups=None,
              network_ids=None, config_drive_files=None, user_data=None,
              port_ids=None, server_group_id=None, availability_zone=None):

        compute_id = self.driver.build(name, amphora_flavor,
                                       image_tag, image_owner,
                                       key_name, sec_groups, network_ids,
                                       config_drive_files, user_data, port_ids,
                                       server_group_id, availability_zone)
        return compute_id

    def delete(self, compute_id):
        self.driver.delete(compute_id)

    def status(self, compute_id):
        return self.driver.status(compute_id)

    def get_amphora(self, compute_id, management_network_id=None):
        return self.driver.get_amphora(compute_id, management_network_id)

    def create_server_group(self, name, policy):
        return self.driver.create_server_group(name, policy)

    def delete_server_group(self, server_group_id):
        self.driver.delete_server_group(server_group_id)

    def attach_network_or_port(self, compute_id, network_id=None,
                               ip_address=None, port_id=None):
        self.driver.attach_network_or_port(compute_id, network_id, ip_address,
                                           port_id)

    def detach_port(self, compute_id, port_id):
        self.driver.detach_port(compute_id, port_id)

    def validate_flavor(self, flavor_id):
        self.driver.validate_flavor(flavor_id)

    def validate_availability_zone(self, availability_zone):
        self.driver.validate_availability_zone(availability_zone)



# Copyright 2014,  Doug Wiegley,  A10 Networks.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import tempfile

from oslo_config import cfg
from oslo_config import fixture as oslo_fixture

import octavia.common.config as config
import octavia.tests.unit.base as base


class TestConfig(base.TestCase):

    def test_sanity(self):
        config.init([])
        config.setup_logging(cfg.CONF)
        # Resetting because this will cause inconsistent errors when run with
        # other tests
        self.addCleanup(cfg.CONF.reset)

    def test_validate_server_certs_key_passphrase(self):
        conf = self.useFixture(oslo_fixture.Config(config.cfg.CONF))
        conf.config(
            group="certificates",
            server_certs_key_passphrase="insecure-key-do-not-use-this-key"
        )

        # Test too short
        self.assertRaises(ValueError, conf.config,
                          group="certificates",
                          server_certs_key_passphrase="short_passphrase")

        # Test too long
        self.assertRaises(
            ValueError, conf.config, group="certificates",
            server_certs_key_passphrase="long-insecure-key-do-not-use-this")

        # Test invalid characters
        self.assertRaises(
            ValueError, conf.config, group="certificates",
            server_certs_key_passphrase="insecure-key-do-not-u$e-this-key")

    def test_active_connection_retry_interval(self):
        conf = self.useFixture(oslo_fixture.Config(config.cfg.CONF))

        # Test new name
        with tempfile.NamedTemporaryFile(mode='w', delete=True) as tmp:
            tmp.write("[haproxy_amphora]\n"
                      "active_connection_retry_interval=4\n")
            tmp.flush()

            conf.set_config_files([tmp.name])

        self.assertEqual(
            4,
            conf.conf.haproxy_amphora.active_connection_retry_interval)

        # Test deprecated name
        with tempfile.NamedTemporaryFile(mode='w', delete=True) as tmp:
            tmp.write("[haproxy_amphora]\n"
                      "active_connection_rety_interval=3\n")
            tmp.flush()

            conf.set_config_files([tmp.name])

        self.assertEqual(
            3,
            conf.conf.haproxy_amphora.active_connection_retry_interval)

    def test_handle_neutron_deprecations(self):
        conf = self.useFixture(oslo_fixture.Config(config.cfg.CONF))

        # The deprecated settings are copied to the new settings
        conf.config(endpoint='my_endpoint',
                    endpoint_type='internal',
                    ca_certificates_file='/path/to/certs',
                    group='neutron')

        config.handle_neutron_deprecations()

        self.assertEqual('my_endpoint', conf.conf.neutron.endpoint_override)
        self.assertEqual(['internal'], conf.conf.neutron.valid_interfaces)
        self.assertEqual('/path/to/certs', conf.conf.neutron.cafile)

    # Test case for https://bugs.launchpad.net/octavia/+bug/2051604
    def test_handle_neutron_deprecations_with_precedence(self):
        conf = self.useFixture(oslo_fixture.Config(config.cfg.CONF))

        # The deprecated settings should not override the new settings when
        # they exist
        conf.config(endpoint='my_old_endpoint',
                    endpoint_type='old_type',
                    ca_certificates_file='/path/to/old_certs',
                    endpoint_override='my_endpoint',
                    valid_interfaces=['internal'],
                    cafile='/path/to/certs',
                    group='neutron')

        config.handle_neutron_deprecations()

        self.assertEqual('my_endpoint', conf.conf.neutron.endpoint_override)
        self.assertEqual(['internal'], conf.conf.neutron.valid_interfaces)
        self.assertEqual('/path/to/certs', conf.conf.neutron.cafile)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#

from oslo_log import log as logging
from taskflow import task

from octavia.common import constants # noqa H306
from octavia.common import context
from octavia.common import rpc

LOG = logging.getLogger(__name__)


class BaseNotificationTask(task.Task):
    event_type = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._rpc_notifier = rpc.get_notifier()

    def execute(self, loadbalancer):
        ctx = context.RequestContext(
            project_id=loadbalancer[constants.PROJECT_ID])
        LOG.debug(f"Sending rpc notification: {self.event_type} "
                  f"{loadbalancer[constants.LOADBALANCER_ID]}")
        self._rpc_notifier.info(
            ctx,
            self.event_type,
            loadbalancer
        )


class SendUpdateNotification(BaseNotificationTask):
    event_type = 'octavia.loadbalancer.update.end'


class SendCreateNotification(BaseNotificationTask):
    event_type = 'octavia.loadbalancer.create.end'


class SendDeleteNotification(BaseNotificationTask):
    event_type = 'octavia.loadbalancer.delete.end'



#    Copyright 2014 Rackspace
#    Copyright 2016 Blue Box, an IBM Company
#    Copyright 2017 Walmart Stores Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_db.sqlalchemy import models
import sqlalchemy as sa
from sqlalchemy.ext import orderinglist
from sqlalchemy import orm
from sqlalchemy.orm import Mapped
from sqlalchemy.orm import validates
from sqlalchemy.sql import func
from sqlalchemy_utils import ScalarListType

from octavia.api.v2.types import amphora
from octavia.api.v2.types import availability_zone_profile
from octavia.api.v2.types import availability_zones
from octavia.api.v2.types import flavor_profile
from octavia.api.v2.types import flavors
from octavia.api.v2.types import health_monitor
from octavia.api.v2.types import l7policy
from octavia.api.v2.types import l7rule
from octavia.api.v2.types import listener
from octavia.api.v2.types import load_balancer
from octavia.api.v2.types import member
from octavia.api.v2.types import pool
from octavia.api.v2.types import quotas
from octavia.common import constants
from octavia.common import data_models
from octavia.db import base_models
from octavia.i18n import _


class ProvisioningStatus(base_models.BASE, base_models.LookupTableMixin):

    __tablename__ = "provisioning_status"


class OperatingStatus(base_models.BASE, base_models.LookupTableMixin):

    __tablename__ = "operating_status"


class Protocol(base_models.BASE, base_models.LookupTableMixin):

    __tablename__ = "protocol"


class Algorithm(base_models.BASE, base_models.LookupTableMixin):

    __tablename__ = "algorithm"


class AmphoraRoles(base_models.BASE, base_models.LookupTableMixin):

    __tablename__ = "amphora_roles"


class LBTopology(base_models.BASE, base_models.LookupTableMixin):

    __tablename__ = "lb_topology"


class SessionPersistenceType(base_models.BASE, base_models.LookupTableMixin):

    __tablename__ = "session_persistence_type"


class HealthMonitorType(base_models.BASE, base_models.LookupTableMixin):

    __tablename__ = "health_monitor_type"


class VRRPAuthMethod(base_models.BASE, base_models.LookupTableMixin):

    __tablename__ = "vrrp_auth_method"


class L7RuleType(base_models.BASE, base_models.LookupTableMixin):

    __tablename__ = "l7rule_type"


class L7RuleCompareType(base_models.BASE, base_models.LookupTableMixin):

    __tablename__ = "l7rule_compare_type"


class L7PolicyAction(base_models.BASE, base_models.LookupTableMixin):

    __tablename__ = "l7policy_action"


class AmphoraBuildSlots(base_models.BASE):

    __tablename__ = "amphora_build_slots"

    id = sa.Column(sa.Integer(), primary_key=True)
    slots_used = sa.Column(sa.Integer())


class AmphoraBuildRequest(base_models.BASE):

    __tablename__ = "amphora_build_request"

    amphora_id = sa.Column(sa.String(36), nullable=True, primary_key=True)
    priority = sa.Column(sa.Integer())
    created_time = sa.Column(sa.DateTime, default=func.now(), nullable=False)
    status = sa.Column(sa.String(16), default='WAITING', nullable=False)


class SessionPersistence(base_models.BASE):

    __data_model__ = data_models.SessionPersistence

    __tablename__ = "session_persistence"

    pool_id = sa.Column(
        sa.String(36),
        sa.ForeignKey("pool.id", name="fk_session_persistence_pool_id"),
        nullable=False,
        primary_key=True)
    type = sa.Column(
        sa.String(36),
        sa.ForeignKey(
            "session_persistence_type.name",
            name="fk_session_persistence_session_persistence_type_name"),
        nullable=False)
    cookie_name = sa.Column(sa.String(255), nullable=True)
    persistence_timeout = sa.Column(sa.Integer(), nullable=True)
    persistence_granularity = sa.Column(sa.String(64), nullable=True)
    pool = orm.relationship("Pool", uselist=False,
                            back_populates="session_persistence")


class ListenerStatistics(base_models.BASE):

    __data_model__ = data_models.ListenerStatistics

    __tablename__ = "listener_statistics"

    listener_id = sa.Column(
        sa.String(36),
        primary_key=True,
        nullable=False)
    amphora_id = sa.Column(
        sa.String(36),
        primary_key=True,
        nullable=False)
    bytes_in = sa.Column(sa.BigInteger, nullable=False)
    bytes_out = sa.Column(sa.BigInteger, nullable=False)
    active_connections = sa.Column(sa.Integer, nullable=False)
    total_connections = sa.Column(sa.BigInteger, nullable=False)
    request_errors = sa.Column(sa.BigInteger, nullable=False)

    @validates('bytes_in', 'bytes_out',
               'active_connections', 'total_connections',
               'request_errors')
    def validate_non_negative_int(self, key, value):
        if value < 0:
            data = {'key': key, 'value': value}
            raise ValueError(_('The %(key)s field can not have '
                               'negative value. '
                               'Current value is %(value)d.') % data)
        return value

    def __iadd__(self, other):
        if isinstance(other, (ListenerStatistics,
                              data_models.ListenerStatistics)):
            self.bytes_in += other.bytes_in
            self.bytes_out += other.bytes_out
            self.request_errors += other.request_errors
            self.total_connections += other.total_connections
        else:
            raise TypeError(  # noqa: O342
                "unsupported operand type(s) for +=: '{}' and '{}'".format(
                    type(self), type(other)))

        return self


class Member(base_models.BASE, base_models.IdMixin, base_models.ProjectMixin,
             models.TimestampMixin, base_models.NameMixin,
             base_models.TagMixin):

    __data_model__ = data_models.Member

    __tablename__ = "member"

    __v2_wsme__ = member.MemberResponse

    __table_args__ = (
        sa.UniqueConstraint('pool_id', 'ip_address', 'protocol_port',
                            name='uq_member_pool_id_address_protocol_port'),
    )

    pool_id = sa.Column(
        sa.String(36),
        sa.ForeignKey("pool.id", name="fk_member_pool_id"),
        nullable=False)
    subnet_id = sa.Column(sa.String(36), nullable=True)
    ip_address = sa.Column('ip_address', sa.String(64), nullable=False)
    protocol_port = sa.Column(sa.Integer, nullable=False)
    weight = sa.Column(sa.Integer, nullable=True)
    backup = sa.Column(sa.Boolean(), nullable=False)
    monitor_address = sa.Column(sa.String(64), nullable=True)
    monitor_port = sa.Column(sa.Integer, nullable=True)
    provisioning_status = sa.Column(
        sa.String(16),
        sa.ForeignKey("provisioning_status.name",
                      name="fk_member_provisioning_status_name"),
        nullable=False)
    operating_status = sa.Column(
        sa.String(16),
        sa.ForeignKey("operating_status.name",
                      name="fk_member_operating_status_name"),
        nullable=False)
    enabled = sa.Column(sa.Boolean(), nullable=False)
    pool = orm.relationship("Pool", back_populates="members")
    vnic_type = sa.Column(sa.String(64), nullable=True)

    _tags = orm.relationship(
        'Tags',
        single_parent=True,
        lazy='subquery',
        cascade='all,delete-orphan',
        primaryjoin='and_(foreign(Tags.resource_id)==Member.id)',
        overlaps='_tags'
    )

    def __str__(self):
        return (f"Member(id={self.id!r}, name={self.name!r}, "
                f"project_id={self.project_id!r}, "
                f"provisioning_status={self.provisioning_status!r}, "
                f"ip_address={self.ip_address!r}, "
                f"protocol_port={self.protocol_port!r}, "
                f"operating_status={self.operating_status!r}, "
                f"weight={self.weight!r}, vnic_type={self.vnic_type!r})")


class HealthMonitor(base_models.BASE, base_models.IdMixin,
                    base_models.ProjectMixin, models.TimestampMixin,
                    base_models.NameMixin, base_models.TagMixin):

    __data_model__ = data_models.HealthMonitor

    __tablename__ = "health_monitor"

    __v2_wsme__ = health_monitor.HealthMonitorResponse

    __table_args__ = (
        sa.UniqueConstraint('pool_id',
                            name='uq_health_monitor_pool'),
    )

    type = sa.Column(
        sa.String(36),
        sa.ForeignKey("health_monitor_type.name",
                      name="fk_health_monitor_health_monitor_type_name"),
        nullable=False)
    pool_id = sa.Column(
        sa.String(36),
        sa.ForeignKey("pool.id", name="fk_health_monitor_pool_id"),
        nullable=False)
    delay = sa.Column(sa.Integer, nullable=False)
    timeout = sa.Column(sa.Integer, nullable=False)
    fall_threshold = sa.Column(sa.Integer, nullable=False)
    rise_threshold = sa.Column(sa.Integer, nullable=False)
    http_method = sa.Column(sa.String(16), nullable=True)
    url_path = sa.Column(sa.String(2048), nullable=True)
    expected_codes = sa.Column(sa.String(64), nullable=True)
    enabled = sa.Column(sa.Boolean, nullable=False)
    pool = orm.relationship("Pool", uselist=False,
                            back_populates="health_monitor")

    provisioning_status = sa.Column(
        sa.String(16),
        sa.ForeignKey("provisioning_status.name",
                      name="fk_health_monitor_provisioning_status_name"),
        nullable=False)
    operating_status = sa.Column(
        sa.String(16),
        sa.ForeignKey("operating_status.name",
                      name="fk_health_monitor_operating_status_name"),
        nullable=False)
    _tags = orm.relationship(
        'Tags',
        single_parent=True,
        lazy='subquery',
        cascade='all,delete-orphan',
        primaryjoin='and_(foreign(Tags.resource_id)==HealthMonitor.id)',
        overlaps='_tags'
    )
    http_version = sa.Column(sa.Float, nullable=True)
    domain_name = sa.Column(sa.String(255), nullable=True)

    def __str__(self):
        return (f"HealthMonitor(id={self.id!r}, name={self.name!r}, "
                f"project_id={self.project_id!r}, type={self.type!r}, "
                f"enabled={self.enabled!r})")


class Pool(base_models.BASE, base_models.IdMixin, base_models.ProjectMixin,
           models.TimestampMixin, base_models.NameMixin, base_models.TagMixin):

    __data_model__ = data_models.Pool

    __tablename__ = "pool"

    __v2_wsme__ = pool.PoolResponse

    description = sa.Column(sa.String(255), nullable=True)
    protocol = sa.Column(
        sa.String(16),
        sa.ForeignKey("protocol.name", name="fk_pool_protocol_name"),
        nullable=False)
    lb_algorithm = sa.Column(
        sa.String(255),
        sa.ForeignKey("algorithm.name", name="fk_pool_algorithm_name"),
        nullable=False)
    provisioning_status = sa.Column(
        sa.String(16),
        sa.ForeignKey("provisioning_status.name",
                      name="fk_pool_provisioning_status_name"),
        nullable=False)
    operating_status = sa.Column(
        sa.String(16),
        sa.ForeignKey("operating_status.name",
                      name="fk_pool_operating_status_name"),
        nullable=False)
    enabled = sa.Column(sa.Boolean, nullable=False)
    load_balancer_id = sa.Column(
        sa.String(36),
        sa.ForeignKey("load_balancer.id", name="fk_pool_load_balancer_id"),
        nullable=True)
    health_monitor = orm.relationship("HealthMonitor", uselist=False,
                                      cascade="delete", back_populates="pool")
    load_balancer = orm.relationship("LoadBalancer", uselist=False,
                                     back_populates="pools")
    members = orm.relationship("Member", uselist=True, cascade="delete",
                               back_populates="pool")
    session_persistence = orm.relationship(
        "SessionPersistence", uselist=False, cascade="delete",
        back_populates="pool")
    _default_listeners = orm.relationship("Listener", uselist=True,
                                          back_populates="default_pool",
                                          cascade_backrefs=False)
    l7policies = orm.relationship("L7Policy", uselist=True,
                                  back_populates="redirect_pool")
    _tags = orm.relationship(
        'Tags',
        single_parent=True,
        lazy='subquery',
        cascade='all,delete-orphan',
        primaryjoin='and_(foreign(Tags.resource_id)==Pool.id)',
        overlaps='_tags'
    )
    tls_certificate_id = sa.Column(sa.String(255), nullable=True)
    ca_tls_certificate_id = sa.Column(sa.String(255), nullable=True)
    crl_container_id = sa.Column(sa.String(255), nullable=True)
    tls_enabled = sa.Column(sa.Boolean, default=False, nullable=False)
    tls_ciphers = sa.Column(sa.String(2048), nullable=True)
    tls_versions = sa.Column(ScalarListType(), nullable=True)
    alpn_protocols = sa.Column(ScalarListType(), nullable=True)

    # This property should be a unique list of any listeners that reference
    # this pool as its default_pool and any listeners referenced by enabled
    # L7Policies with at least one l7rule which also reference this pool. The
    # intent is that pool.listeners should be a unique list of listeners
    # *actually* using the pool.
    @property
    def listeners(self):
        _listeners = self._default_listeners[:]
        _l_ids = [li.id for li in _listeners]
        l7_listeners = [p.listener for p in self.l7policies
                        if len(p.l7rules) > 0 and p.enabled is True]
        for li in l7_listeners:
            if li.id not in _l_ids:
                _listeners.append(li)
                _l_ids.append(li.id)
        return _listeners

    def __str__(self):
        return (f"Pool(id={self.id!r}, name={self.name!r}, "
                f"project_id={self.project_id!r}, "
                f"provisioning_status={self.provisioning_status!r}, "
                f"protocol={self.protocol!r}, "
                f"lb_algorithm={self.lb_algorithm!r}, "
                f"enabled={self.enabled!r})")


class LoadBalancer(base_models.BASE, base_models.IdMixin,
                   base_models.ProjectMixin, models.TimestampMixin,
                   base_models.NameMixin, base_models.TagMixin):

    __data_model__ = data_models.LoadBalancer

    __tablename__ = "load_balancer"

    __v2_wsme__ = load_balancer.LoadBalancerResponse

    description = sa.Column(sa.String(255), nullable=True)
    provisioning_status = sa.Column(
        sa.String(16),
        sa.ForeignKey("provisioning_status.name",
                      name="fk_load_balancer_provisioning_status_name"),
        nullable=False)
    operating_status = sa.Column(
        sa.String(16),
        sa.ForeignKey("operating_status.name",
                      name="fk_load_balancer_operating_status_name"),
        nullable=False)
    topology = sa.Column(
        sa.String(36),
        sa.ForeignKey("lb_topology.name", name="fk_lb_topology_name"),
        nullable=True)
    enabled = sa.Column(sa.Boolean, nullable=False)
    amphorae = orm.relationship("Amphora", uselist=True,
                                back_populates="load_balancer")
    server_group_id = sa.Column(sa.String(36), nullable=True)
    provider = sa.Column(sa.String(64), nullable=True)
    vip = orm.relationship('Vip', cascade='delete', uselist=False,
                           backref=orm.backref('load_balancer', uselist=False))
    additional_vips = orm.relationship(
        'AdditionalVip', cascade='delete', uselist=True,
        backref=orm.backref('load_balancer', uselist=False))
    pools = orm.relationship('Pool', cascade='delete', uselist=True,
                             back_populates="load_balancer")
    listeners = orm.relationship('Listener', cascade='delete', uselist=True,
                                 back_populates='load_balancer')
    _tags = orm.relationship(
        'Tags',
        single_parent=True,
        lazy='subquery',
        cascade='all,delete-orphan',
        primaryjoin='and_(foreign(Tags.resource_id)==LoadBalancer.id)',
        overlaps='_tags'
    )
    flavor_id = sa.Column(
        sa.String(36),
        sa.ForeignKey("flavor.id", name="fk_lb_flavor_id"), nullable=True)
    availability_zone = sa.Column(
        sa.String(255),
        sa.ForeignKey("availability_zone.name",
                      name="fk_load_balancer_availability_zone_name"),
        nullable=True)
    flavor: Mapped["Flavor"] = orm.relationship("Flavor")

    def __str__(self):
        return (f"LoadBalancer(id={self.id!r}, name={self.name!r}, "
                f"project_id={self.project_id!r}, vip={self.vip!r}, "
                f"provisioning_status={self.provisioning_status!r}, "
                f"operating_status={self.operating_status!r}, "
                f"provider={self.provider!r})")


class VRRPGroup(base_models.BASE):

    __data_model__ = data_models.VRRPGroup

    __tablename__ = "vrrp_group"

    load_balancer_id = sa.Column(
        sa.String(36),
        sa.ForeignKey("load_balancer.id",
                      name="fk_vrrp_group_load_balancer_id"),
        nullable=False, primary_key=True)

    vrrp_group_name = sa.Column(sa.String(36), nullable=True)
    vrrp_auth_type = sa.Column(sa.String(16), sa.ForeignKey(
        "vrrp_auth_method.name",
        name="fk_load_balancer_vrrp_auth_method_name"))
    vrrp_auth_pass = sa.Column(sa.String(36), nullable=True)
    advert_int = sa.Column(sa.Integer(), nullable=True)
    load_balancer = orm.relationship("LoadBalancer", uselist=False,
                                     backref=orm.backref("vrrp_group",
                                                         uselist=False,
                                                         cascade="delete"))


class Vip(base_models.BASE):

    __data_model__ = data_models.Vip

    __tablename__ = "vip"

    load_balancer_id = sa.Column(
        sa.String(36),
        sa.ForeignKey("load_balancer.id",
                      name="fk_vip_load_balancer_id"),
        nullable=False, primary_key=True)
    ip_address = sa.Column(sa.String(64), nullable=True)
    port_id = sa.Column(sa.String(36), nullable=True)
    subnet_id = sa.Column(sa.String(36), nullable=True)
    network_id = sa.Column(sa.String(36), nullable=True)
    qos_policy_id = sa.Column(sa.String(36), nullable=True)
    octavia_owned = sa.Column(sa.Boolean(), nullable=True)
    vnic_type = sa.Column(sa.String(64), nullable=True)

    sgs = orm.relationship(
        "VipSecurityGroup", cascade="all,delete-orphan",
        uselist=True, backref=orm.backref("vip", uselist=False))

    @property
    def sg_ids(self) -> list[str]:
        return [sg.sg_id for sg in self.sgs]


class AdditionalVip(base_models.BASE):

    __data_model__ = data_models.AdditionalVip

    __tablename__ = "additional_vip"

    __table_args__ = (
        sa.PrimaryKeyConstraint('load_balancer_id', 'subnet_id',
                                name='pk_add_vip_load_balancer_subnet'),
    )

    load_balancer_id = sa.Column(
        sa.String(36),
        sa.ForeignKey("load_balancer.id",
                      name="fk_add_vip_load_balancer_id"),
        nullable=False, index=True)
    ip_address = sa.Column(sa.String(64), nullable=True)
    port_id = sa.Column(sa.String(36), nullable=True)
    subnet_id = sa.Column(sa.String(36), nullable=True)
    network_id = sa.Column(sa.String(36), nullable=True)


class Listener(base_models.BASE, base_models.IdMixin,
               base_models.ProjectMixin, models.TimestampMixin,
               base_models.NameMixin, base_models.TagMixin):

    __data_model__ = data_models.Listener

    __tablename__ = "listener"

    __v2_wsme__ = listener.ListenerResponse

    __table_args__ = (
        sa.UniqueConstraint(
            'load_balancer_id', 'protocol', 'protocol_port',
            name='uq_listener_load_balancer_id_protocol_port'),
    )

    description = sa.Column(sa.String(255), nullable=True)
    protocol = sa.Column(
        sa.String(16),
        sa.ForeignKey("protocol.name", name="fk_listener_protocol_name"),
        nullable=False)
    protocol_port = sa.Column(sa.Integer(), nullable=False)
    connection_limit = sa.Column(sa.Integer, nullable=True)
    load_balancer_id = sa.Column(
        sa.String(36),
        sa.ForeignKey("load_balancer.id", name="fk_listener_load_balancer_id"),
        nullable=True)
    tls_certificate_id = sa.Column(sa.String(255), nullable=True)
    default_pool_id = sa.Column(
        sa.String(36),
        sa.ForeignKey("pool.id", name="fk_listener_pool_id"),
        nullable=True)
    provisioning_status = sa.Column(
        sa.String(16),
        sa.ForeignKey("provisioning_status.name",
                      name="fk_listener_provisioning_status_name"),
        nullable=False)
    operating_status = sa.Column(
        sa.String(16),
        sa.ForeignKey("operating_status.name",
                      name="fk_listener_operating_status_name"),
        nullable=False)
    enabled = sa.Column(sa.Boolean(), nullable=False)
    load_balancer = orm.relationship("LoadBalancer", uselist=False,
                                     back_populates="listeners")
    default_pool = orm.relationship("Pool", uselist=False,
                                    back_populates="_default_listeners",
                                    cascade_backrefs=False)
    sni_containers = orm.relationship(
        'SNI', cascade='all,delete-orphan',
        uselist=True, backref=orm.backref('listener', uselist=False))

    l7policies = orm.relationship(
        'L7Policy', uselist=True, order_by='L7Policy.position',
        collection_class=orderinglist.ordering_list('position', count_from=1),
        cascade='delete', back_populates='listener')

    peer_port = sa.Column(sa.Integer(), nullable=True)
    insert_headers = sa.Column(sa.PickleType())
    timeout_client_data = sa.Column(sa.Integer, nullable=True)
    timeout_member_connect = sa.Column(sa.Integer, nullable=True)
    timeout_member_data = sa.Column(sa.Integer, nullable=True)
    timeout_tcp_inspect = sa.Column(sa.Integer, nullable=True)
    client_ca_tls_certificate_id = sa.Column(sa.String(255), nullable=True)
    client_authentication = sa.Column(
        sa.String(10),
        sa.ForeignKey("client_authentication_mode.name",
                      name="fk_listener_client_authentication_mode_name"),
        nullable=False, default=constants.CLIENT_AUTH_NONE)
    client_crl_container_id = sa.Column(sa.String(255), nullable=True)
    tls_ciphers = sa.Column(sa.String(2048), nullable=True)
    tls_versions = sa.Column(ScalarListType(), nullable=True)
    alpn_protocols = sa.Column(ScalarListType(), nullable=True)
    hsts_max_age = sa.Column(sa.Integer, nullable=True)
    hsts_include_subdomains = sa.Column(sa.Boolean, nullable=True)
    hsts_preload = sa.Column(sa.Boolean, nullable=True)

    _tags = orm.relationship(
        'Tags',
        single_parent=True,
        lazy='subquery',
        cascade='all,delete-orphan',
        primaryjoin='and_(foreign(Tags.resource_id)==Listener.id)',
        overlaps='_tags'
    )

    # This property should be a unique list of the default_pool and anything
    # referenced by enabled L7Policies with at least one rule that also
    # reference this listener. The intent is that listener.pools should be a
    # unique list of pools this listener is *actually* using.
    @property
    def pools(self):
        _pools = []
        _p_ids = []
        if self.default_pool:
            _pools.append(self.default_pool)
            _p_ids.append(self.default_pool.id)
        l7_pools = [p.redirect_pool for p in self.l7policies
                    if p.redirect_pool is not None and len(p.l7rules) > 0 and
                    p.enabled is True]
        for p in l7_pools:
            if p.id not in _p_ids:
                _pools.append(p)
                _p_ids.append(p.id)
        return _pools

    allowed_cidrs = orm.relationship(
        'ListenerCidr', cascade='all,delete-orphan',
        uselist=True, backref=orm.backref('listener', uselist=False))

    def __str__(self):
        return (f"Listener(id={self.id!r}, "
                f"default_pool={self.default_pool!r}, name={self.name!r}, "
                f"project_id={self.project_id!r}, protocol={self.protocol!r}, "
                f"protocol_port={self.protocol_port!r}, "
                f"enabled={self.enabled!r})")


class SNI(base_models.BASE):

    __data_model__ = data_models.SNI

    __tablename__ = "sni"
    __table_args__ = (
        sa.PrimaryKeyConstraint('listener_id', 'tls_container_id'),
    )
    listener_id = sa.Column(
        sa.String(36),
        sa.ForeignKey("listener.id", name="fk_sni_listener_id"),
        nullable=False)
    tls_container_id = sa.Column(sa.String(128), nullable=False)
    position = sa.Column(sa.Integer(), nullable=True)


class Amphora(base_models.BASE, base_models.IdMixin, models.TimestampMixin):

    __data_model__ = data_models.Amphora

    __tablename__ = "amphora"

    __v2_wsme__ = amphora.AmphoraResponse

    load_balancer_id = sa.Column(
        sa.String(36), sa.ForeignKey("load_balancer.id",
                                     name="fk_amphora_load_balancer_id"),
        nullable=True)
    compute_id = sa.Column(sa.String(36), nullable=True)
    lb_network_ip = sa.Column(sa.String(64), nullable=True)
    vrrp_ip = sa.Column(sa.String(64), nullable=True)
    ha_ip = sa.Column(sa.String(64), nullable=True)
    vrrp_port_id = sa.Column(sa.String(36), nullable=True)
    ha_port_id = sa.Column(sa.String(36), nullable=True)
    cert_expiration = sa.Column(sa.DateTime(timezone=True), default=None,
                                nullable=True)
    cert_busy = sa.Column(sa.Boolean(), default=False, nullable=False)

    role = sa.Column(
        sa.String(36),
        sa.ForeignKey("amphora_roles.name", name="fk_amphora_roles_name"),
        nullable=True)
    status = sa.Column(
        sa.String(36),
        sa.ForeignKey("provisioning_status.name",
                      name="fk_container_provisioning_status_name"))
    vrrp_interface = sa.Column(sa.String(16), nullable=True)
    vrrp_id = sa.Column(sa.Integer(), nullable=True)
    vrrp_priority = sa.Column(sa.Integer(), nullable=True)
    cached_zone = sa.Column(sa.String(255), nullable=True)
    image_id = sa.Column(sa.String(36), nullable=True)
    load_balancer = orm.relationship("LoadBalancer", uselist=False,
                                     back_populates='amphorae')
    compute_flavor = sa.Column(sa.String(255), nullable=True)

    def __str__(self):
        return (f"Amphora(id={self.id!r}, load_balancer_id="
                f"{self.load_balancer_id!r}, status={self.status!r}, "
                f"role={self.role!r}, lb_network_ip={self.lb_network_ip!r}, "
                f"vrrp_ip={self.vrrp_ip!r})")


class AmphoraHealth(base_models.BASE):
    __data_model__ = data_models.AmphoraHealth
    __tablename__ = "amphora_health"

    amphora_id = sa.Column(
        sa.String(36), nullable=False, primary_key=True)
    last_update = sa.Column(sa.DateTime, default=func.now(),
                            nullable=False)

    busy = sa.Column(sa.Boolean(), default=False, nullable=False)


class L7Rule(base_models.BASE, base_models.IdMixin, base_models.ProjectMixin,
             models.TimestampMixin, base_models.TagMixin):

    __data_model__ = data_models.L7Rule

    __tablename__ = "l7rule"

    __v2_wsme__ = l7rule.L7RuleResponse

    l7policy_id = sa.Column(
        sa.String(36),
        sa.ForeignKey("l7policy.id", name="fk_l7rule_l7policy_id"),
        nullable=False)
    type = sa.Column(
        sa.String(36),
        sa.ForeignKey(
            "l7rule_type.name",
            name="fk_l7rule_l7rule_type_name"),
        nullable=False)
    compare_type = sa.Column(
        sa.String(36),
        sa.ForeignKey(
            "l7rule_compare_type.name",
            name="fk_l7rule_l7rule_compare_type_name"),
        nullable=False)
    key = sa.Column(sa.String(255), nullable=True)
    value = sa.Column(sa.String(255), nullable=False)
    invert = sa.Column(sa.Boolean(), default=False, nullable=False)
    enabled = sa.Column(sa.Boolean(), nullable=False)
    l7policy = orm.relationship("L7Policy", uselist=False,
                                back_populates="l7rules")
    provisioning_status = sa.Column(
        sa.String(16),
        sa.ForeignKey("provisioning_status.name",
                      name="fk_l7rule_provisioning_status_name"),
        nullable=False)
    operating_status = sa.Column(
        sa.String(16),
        sa.ForeignKey("operating_status.name",
                      name="fk_l7rule_operating_status_name"),
        nullable=False)
    _tags = orm.relationship(
        'Tags',
        single_parent=True,
        lazy='subquery',
        cascade='all,delete-orphan',
        primaryjoin='and_(foreign(Tags.resource_id)==L7Rule.id)',
        overlaps='_tags'
    )

    def __str__(self):
        return (f"L7Rule(id={self.id!r}, project_id={self.project_id!r}, "
                f"provisioning_status={self.provisioning_status!r}, "
                f"type={self.type!r}, key={self.key!r}, value={self.value!r}, "
                f"invert={self.invert!r}, enabled={self.enabled!r})")


class L7Policy(base_models.BASE, base_models.IdMixin, base_models.ProjectMixin,
               models.TimestampMixin, base_models.NameMixin,
               base_models.TagMixin):

    __data_model__ = data_models.L7Policy

    __tablename__ = "l7policy"

    __v2_wsme__ = l7policy.L7PolicyResponse

    description = sa.Column(sa.String(255), nullable=True)
    listener_id = sa.Column(
        sa.String(36),
        sa.ForeignKey("listener.id", name="fk_l7policy_listener_id"),
        nullable=False)
    action = sa.Column(
        sa.String(36),
        sa.ForeignKey(
            "l7policy_action.name",
            name="fk_l7policy_l7policy_action_name"),
        nullable=False)
    redirect_pool_id = sa.Column(
        sa.String(36),
        sa.ForeignKey("pool.id", name="fk_l7policy_pool_id"),
        nullable=True)
    redirect_url = sa.Column(
        sa.String(255),
        nullable=True)
    redirect_prefix = sa.Column(
        sa.String(255),
        nullable=True)
    redirect_http_code = sa.Column(sa.Integer, nullable=True)
    position = sa.Column(sa.Integer, nullable=False)
    enabled = sa.Column(sa.Boolean(), nullable=False)
    listener = orm.relationship("Listener", uselist=False,
                                back_populates="l7policies")
    redirect_pool = orm.relationship("Pool", uselist=False,
                                     back_populates="l7policies")
    l7rules = orm.relationship("L7Rule", uselist=True, cascade="delete",
                               back_populates="l7policy")
    provisioning_status = sa.Column(
        sa.String(16),
        sa.ForeignKey("provisioning_status.name",
                      name="fk_l7policy_provisioning_status_name"),
        nullable=False)
    operating_status = sa.Column(
        sa.String(16),
        sa.ForeignKey("operating_status.name",
                      name="fk_l7policy_operating_status_name"),
        nullable=False)
    _tags = orm.relationship(
        'Tags',
        single_parent=True,
        lazy='subquery',
        cascade='all,delete-orphan',
        primaryjoin='and_(foreign(Tags.resource_id)==L7Policy.id)',
        overlaps='_tags'
    )

    def __str__(self):
        return (f"L7Policy(id={self.id!r}, name={self.name!r}, "
                f"project_id={self.project_id!r}, "
                f"provisioning_status={self.provisioning_status!r}, "
                f"action={self.action!r}, position={self.position!r}, "
                f"enabled={self.enabled!r})")


class Quotas(base_models.BASE):

    __data_model__ = data_models.Quotas

    __tablename__ = "quotas"

    __v2_wsme__ = quotas.QuotaAllBase

    project_id = sa.Column(sa.String(36), primary_key=True)
    health_monitor = sa.Column(sa.Integer(), nullable=True)
    listener = sa.Column(sa.Integer(), nullable=True)
    load_balancer = sa.Column(sa.Integer(), nullable=True)
    member = sa.Column(sa.Integer(), nullable=True)
    pool = sa.Column(sa.Integer(), nullable=True)
    l7policy = sa.Column(sa.Integer(), nullable=True)
    l7rule = sa.Column(sa.Integer(), nullable=True)
    in_use_health_monitor = sa.Column(sa.Integer(), nullable=True)
    in_use_listener = sa.Column(sa.Integer(), nullable=True)
    in_use_load_balancer = sa.Column(sa.Integer(), nullable=True)
    in_use_member = sa.Column(sa.Integer(), nullable=True)
    in_use_pool = sa.Column(sa.Integer(), nullable=True)
    in_use_l7policy = sa.Column(sa.Integer(), nullable=True)
    in_use_l7rule = sa.Column(sa.Integer(), nullable=True)

    def __str__(self):
        return (f"Quotas(project_id={self.project_id!r}, "
                f"load_balancer={self.load_balancer!r}, "
                f"listener={self.listener!r}, pool={self.pool!r}, "
                f"health_monitor={self.health_monitor!r}, "
                f"member={self.member!r}, l7policy={self.l7policy!r}, "
                f"l7rule={self.l7rule!r})")


class FlavorProfile(base_models.BASE, base_models.IdMixin,
                    base_models.NameMixin):

    __data_model__ = data_models.FlavorProfile

    __tablename__ = "flavor_profile"

    __v2_wsme__ = flavor_profile.FlavorProfileResponse

    provider_name = sa.Column(sa.String(255), nullable=False)
    flavor_data = sa.Column(sa.String(4096), nullable=False)


class Flavor(base_models.BASE,
             base_models.IdMixin,
             base_models.NameMixin):

    __data_model__ = data_models.Flavor

    __tablename__ = "flavor"

    __v2_wsme__ = flavors.FlavorResponse

    __table_args__ = (
        sa.UniqueConstraint('name',
                            name='uq_flavor_name'),
    )

    description = sa.Column(sa.String(255), nullable=True)
    enabled = sa.Column(sa.Boolean(), nullable=False)
    flavor_profile_id = sa.Column(
        sa.String(36),
        sa.ForeignKey("flavor_profile.id",
                      name="fk_flavor_flavor_profile_id"),
        nullable=False)
    flavor_profile: Mapped["FlavorProfile"] = orm.relationship("FlavorProfile")


class AvailabilityZoneProfile(base_models.BASE, base_models.IdMixin,
                              base_models.NameMixin):

    __data_model__ = data_models.AvailabilityZoneProfile

    __tablename__ = "availability_zone_profile"

    __v2_wsme__ = availability_zone_profile.AvailabilityZoneProfileResponse

    provider_name = sa.Column(sa.String(255), nullable=False)
    availability_zone_data = sa.Column(sa.String(4096), nullable=False)


class AvailabilityZone(base_models.BASE,
                       base_models.NameMixin):

    __data_model__ = data_models.AvailabilityZone

    __tablename__ = "availability_zone"

    __v2_wsme__ = availability_zones.AvailabilityZoneResponse

    __table_args__ = (
        sa.PrimaryKeyConstraint('name'),
    )

    description = sa.Column(sa.String(255), nullable=True)
    enabled = sa.Column(sa.Boolean(), nullable=False)
    availability_zone_profile_id = sa.Column(
        sa.String(36),
        sa.ForeignKey("availability_zone_profile.id",
                      name="fk_az_az_profile_id"),
        nullable=False)
    availability_zone_profile: Mapped["AvailabilityZoneProfile"] = (
        orm.relationship("AvailabilityZoneProfile"))


class ClientAuthenticationMode(base_models.BASE):

    __tablename__ = "client_authentication_mode"

    name = sa.Column(sa.String(10), primary_key=True, nullable=False)


class ListenerCidr(base_models.BASE):

    __data_model__ = data_models.ListenerCidr

    __tablename__ = "listener_cidr"
    __table_args__ = (
        sa.PrimaryKeyConstraint('listener_id', 'cidr'),
    )

    listener_id = sa.Column(
        sa.String(36),
        sa.ForeignKey("listener.id", name="fk_listener_cidr_listener_id"),
        nullable=False)
    cidr = sa.Column(sa.String(64), nullable=False)


class VipSecurityGroup(base_models.BASE):

    __data_model__ = data_models.VipSecurityGroup

    __tablename__ = "vip_security_group"
    __table_args__ = (
        sa.PrimaryKeyConstraint('load_balancer_id', 'sg_id'),
    )

    load_balancer_id = sa.Column(
        sa.String(36),
        sa.ForeignKey("vip.load_balancer_id", name="fk_vip_sg_vip_lb_id"),
        nullable=False)
    sg_id = sa.Column(sa.String(64), nullable=False)


class AmphoraMemberPort(base_models.BASE, models.TimestampMixin):

    __data_model__ = data_models.AmphoraMemberPort

    __tablename__ = "amphora_member_port"

    port_id = sa.Column(
        sa.String(36),
        primary_key=True)
    amphora_id = sa.Column(
        sa.String(36),
        sa.ForeignKey("amphora.id", name="fk_member_port_amphora_id"),
        nullable=False)
    network_id = sa.Column(
        sa.String(36))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import versionutils
from oslo_policy import policy

from octavia.common import constants

# Octavia specific Advanced RBAC rules

# The default is to not allow access unless the auth_strategy is 'noauth'.
# Users must be a member of one of the following roles to have access to
# the load-balancer API:
#
# role:load-balancer_observer
#     User has access to load-balancer read-only APIs
# role:load-balancer_global_observer
#     User has access to load-balancer read-only APIs including resources
#     owned by others.
# role:load-balancer_member
#     User has access to load-balancer read and write APIs
# role:load-balancer_admin
#     User is considered an admin for all load-balancer APIs including
#     resources owned by others.

deprecated_context_is_admin = policy.DeprecatedRule(
    name='context_is_admin',
    check_str='role:admin or '
              'role:load-balancer_admin',
    deprecated_reason=constants.RBAC_ROLES_DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY,
)

# Note: 'is_admin:True' is a policy rule that takes into account the
# auth_strategy == noauth configuration setting.
# It is equivalent to 'rule:context_is_admin or {auth_strategy == noauth}'

deprecated_admin = policy.DeprecatedRule(
    name='load-balancer:admin',
    check_str='is_admin:True or '
              'role:admin or '
              'role:load-balancer_admin',
    deprecated_reason=constants.RBAC_ROLES_DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY,
)

deprecated_global_observer = policy.DeprecatedRule(
    name='load-balancer:global_observer',
    check_str='role:load-balancer_global_observer',
    deprecated_reason=constants.RBAC_ROLES_DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY,
)

deprecated_member_and_owner = policy.DeprecatedRule(
    name='load-balancer:member_and_owner',
    check_str='role:load-balancer_member and '
              'rule:load-balancer:owner',
    deprecated_reason=constants.RBAC_ROLES_DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY,
)

deprecated_observer_and_owner = policy.DeprecatedRule(
    name='load-balancer:observer_and_owner',
    check_str='role:load-balancer_observer and '
              'rule:load-balancer:owner',
    deprecated_reason=constants.RBAC_ROLES_DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY,
)

deprecated_quota_admin = policy.DeprecatedRule(
    name='load-balancer:quota-admin',
    check_str='role:load-balancer_quota_admin',
    deprecated_reason=constants.RBAC_ROLES_DEPRECATED_REASON,
    deprecated_since=versionutils.deprecated.WALLABY,
)

rules = [
    policy.RuleDefault(
        name='load-balancer:owner',
        check_str='project_id:%(project_id)s',
        scope_types=[constants.RBAC_SCOPE_PROJECT]),
]


def list_rules():
    return rules



#    Copyright 2014 Rackspace
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import time

from sqlalchemy.sql.expression import select

from oslo_config import cfg
from oslo_db.sqlalchemy import enginefacade
from oslo_log import log as logging

LOG = logging.getLogger(__name__)
_FACADE = None


def _create_facade_lazily():
    global _FACADE
    if _FACADE is None:
        _FACADE = True
        enginefacade.configure(sqlite_fk=True, expire_on_commit=True)


def _get_transaction_context(reader=False):
    _create_facade_lazily()
    # TODO(gthiemonge) Create and use new functions to get read-only sessions
    if reader:
        context = enginefacade.reader
    else:
        context = enginefacade.writer
    return context


def _get_sessionmaker(reader=False):
    context = _get_transaction_context(reader)
    return context.get_sessionmaker()


def get_engine():
    context = _get_transaction_context()
    return context.get_engine()


def get_session():
    """Helper method to grab session."""
    return _get_sessionmaker()()


def session():
    return _get_sessionmaker()


def wait_for_connection(exit_event):
    """Helper method to wait for DB connection"""
    down = True
    while down and not exit_event.is_set():
        try:
            LOG.debug('Trying to re-establish connection to database.')
            get_engine().scalar(select([1]))
            down = False
            LOG.debug('Connection to database re-established.')
        except Exception:
            retry_interval = cfg.CONF.database.retry_interval
            LOG.exception('Connection to database failed. Retrying in %s '
                          'seconds.', retry_interval)
            time.sleep(retry_interval)



#    Copyright 2022 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from unittest import mock

from oslo_config import cfg
from oslo_config import fixture as oslo_fixture
import pecan

from octavia.api import config as pconfig
from octavia.common import constants
from octavia.tests.functional.db import base as base_db_test


class TestContentTypes(base_db_test.OctaviaDBTestBase):

    def setUp(self):
        super().setUp()

        self.conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        # Mock log_opt_values, it prevents the dump of the configuration
        # with LOG.info for each test. It saves a lot of time when running
        # the functional tests.
        self.conf.conf.log_opt_values = mock.MagicMock()

        # Note: we need to set argv=() to stop the wsgi setup_app from
        # pulling in the testing tool sys.argv
        self.app = pecan.testing.load_test_app({'app': pconfig.app,
                                                'wsme': pconfig.wsme},
                                               argv=())

        def reset_pecan():
            pecan.set_config({}, overwrite=True)

        self.addCleanup(reset_pecan)

        self.test_url = '/'

    def test_no_accept_header(self):
        response = self.app.get(self.test_url, status=200, expect_errors=False)
        self.assertEqual(200, response.status_code)
        self.assertEqual(constants.APPLICATION_JSON, response.content_type)

    # TODO(johnsom) Testing for an empty string is a workaround for an
    #               openstacksdk bug present up to the initial
    #               antelope release of openstacksdk. This means the
    #               octavia dashboard would also be impacted.
    #               This test should change to a 406 error once the workaround
    #               is removed.
    # See: https://review.opendev.org/c/openstack/openstacksdk/+/876669
    def test_empty_accept_header(self):
        response = self.app.get(
            self.test_url, status=200, expect_errors=False,
            headers={constants.ACCEPT: ''})
        self.assertEqual(200, response.status_code)
        self.assertEqual(constants.APPLICATION_JSON, response.content_type)

    # Note: webob will treat invalid content types as no accept header provided
    def test_bogus_accept_header(self):
        response = self.app.get(
            self.test_url, status=200, expect_errors=False,
            headers={constants.ACCEPT: 'bogus'})
        self.assertEqual(200, response.status_code)
        self.assertEqual(constants.APPLICATION_JSON, response.content_type)

    def test_valid_accept_header(self):
        response = self.app.get(
            self.test_url, status=200, expect_errors=False,
            headers={constants.ACCEPT: constants.APPLICATION_JSON})
        self.assertEqual(200, response.status_code)
        self.assertEqual(constants.APPLICATION_JSON, response.content_type)

    def test_valid_mixed_accept_header(self):
        response = self.app.get(
            self.test_url, status=200, expect_errors=False,
            headers={constants.ACCEPT:
                     'text/html,' + constants.APPLICATION_JSON})
        self.assertEqual(200, response.status_code)
        self.assertEqual(constants.APPLICATION_JSON, response.content_type)

    def test_wildcard_accept_header(self):
        response = self.app.get(
            self.test_url, status=200, expect_errors=False,
            headers={constants.ACCEPT: '*/*'})
        self.assertEqual(200, response.status_code)
        self.assertEqual(constants.APPLICATION_JSON, response.content_type)

    def test_json_wildcard_accept_header(self):
        response = self.app.get(
            self.test_url, status=200, expect_errors=False,
            headers={constants.ACCEPT: constants.APPLICATION_JSON + ', */*'})
        self.assertEqual(200, response.status_code)
        self.assertEqual(constants.APPLICATION_JSON, response.content_type)

    def test_json_plain_wildcard_accept_header(self):
        response = self.app.get(
            self.test_url, status=200, expect_errors=False,
            headers={constants.ACCEPT: constants.APPLICATION_JSON +
                     ', text/plain, */*'})
        self.assertEqual(200, response.status_code)
        self.assertEqual(constants.APPLICATION_JSON, response.content_type)

    def test_wildcard_mixed_accept_header(self):
        response = self.app.get(
            self.test_url, status=200, expect_errors=False,
            headers={constants.ACCEPT:
                     'text/html,*/*'})
        self.assertEqual(200, response.status_code)
        self.assertEqual(constants.APPLICATION_JSON, response.content_type)

    def test_valid_mixed_weighted_accept_header(self):
        response = self.app.get(
            self.test_url, status=200, expect_errors=False,
            headers={constants.ACCEPT:
                     'text/html,' + constants.APPLICATION_JSON + ';q=0.8'})
        self.assertEqual(200, response.status_code)
        self.assertEqual(constants.APPLICATION_JSON, response.content_type)

    def test_invalid_accept_header(self):
        response = self.app.get(
            self.test_url, status=406, expect_errors=False,
            headers={constants.ACCEPT: 'application/xml'})
        self.assertEqual(406, response.status_code)
        self.assertEqual(constants.APPLICATION_JSON, response.content_type)
        self.assertEqual(406, response.json[constants.CODE])
        self.assertEqual('Not Acceptable', response.json[constants.TITLE])
        self.assertEqual('Only content type application/json is accepted.',
                         response.json[constants.DESCRIPTION])

    def test_invalid_mixed_accept_header(self):
        response = self.app.get(
            self.test_url, status=406, expect_errors=False,
            headers={constants.ACCEPT: 'application/xml,text/html'})
        self.assertEqual(406, response.status_code)
        self.assertEqual(constants.APPLICATION_JSON, response.content_type)
        self.assertEqual(406, response.json[constants.CODE])
        self.assertEqual('Not Acceptable', response.json[constants.TITLE])
        self.assertEqual('Only content type application/json is accepted.',
                         response.json[constants.DESCRIPTION])



# Copyright 2015 Hewlett-Packard Development Company, L.P.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import typing

from oslo_log import log as logging
from oslo_utils import uuidutils
from sqlalchemy import Column
from sqlalchemy import create_engine
from sqlalchemy import delete
from sqlalchemy import event
from sqlalchemy import insert
from sqlalchemy import MetaData
from sqlalchemy import select
from sqlalchemy import String
from sqlalchemy import Table

from octavia.common import constants
from octavia.common import data_models
from octavia.network import base as driver_base
from octavia.network import data_models as network_models

if typing.TYPE_CHECKING:
    from octavia.common import context

LOG = logging.getLogger(__name__)

_NOOP_MANAGER_VARS = {
    'networks': {},
    'subnets': {},
    'ports': {},
    'interfaces': {},
    'current_network': None
}


class NoopManager:

    def __init__(self):
        super().__init__()
        self.networkconfigconfig = {}
        self._qos_extension_enabled = True

        # The controller worker can run with multiple processes, so we
        # need to have a persistent and shared store for the no-op driver.
        self.engine = create_engine('sqlite:////tmp/octavia-network-noop.db')

        # TODO(johnsom) work around pysqlite locking issues per:
        # https://github.com/sqlalchemy/sqlalchemy/discussions/12330
        @event.listens_for(self.engine, "connect")
        def do_connect(dbapi_connection, connection_record):
            # disable pysqlite's emitting of the BEGIN statement entirely.
            # also stops it from emitting COMMIT before any DDL.
            dbapi_connection.isolation_level = None

        @event.listens_for(self.engine, "begin")
        def do_begin(conn):
            conn.exec_driver_sql("BEGIN EXCLUSIVE")

        metadata_obj = MetaData()

        self.interfaces_table = Table(
            'interfaces',
            metadata_obj,
            Column('port_id', String(36)),
            Column('network_id', String(36)),
            Column('compute_id', String(36)),
            Column('vnic_type', String(6)))
        self.fixed_ips_table = Table(
            'fixed_ips',
            metadata_obj,
            Column('port_id', String(36)),
            Column('subnet_id', String(36)),
            Column('ip_address', String(64)))
        metadata_obj.create_all(self.engine)

    def allocate_vip(self, loadbalancer):
        LOG.debug("Network %s no-op, allocate_vip loadbalancer %s",
                  self.__class__.__name__, loadbalancer)
        self.networkconfigconfig[loadbalancer.id] = (
            loadbalancer, 'allocate_vip')
        subnet_id = uuidutils.generate_uuid()
        network_id = uuidutils.generate_uuid()
        port_id = uuidutils.generate_uuid()
        ip_address = '198.51.100.1'
        if loadbalancer.vip:
            subnet_id = loadbalancer.vip.subnet_id or subnet_id
            network_id = loadbalancer.vip.network_id or network_id
            port_id = loadbalancer.vip.port_id or port_id
            ip_address = loadbalancer.vip.ip_address or ip_address
        return_vip = data_models.Vip(ip_address=ip_address,
                                     subnet_id=subnet_id,
                                     network_id=network_id,
                                     port_id=port_id,
                                     load_balancer_id=loadbalancer.id)
        additional_vips = [
            data_models.AdditionalVip(
                ip_address=add_vip.ip_address,
                subnet_id=add_vip.subnet_id,
                network_id=network_id,
                port_id=port_id,
                load_balancer=loadbalancer,
                load_balancer_id=loadbalancer.id)
            for add_vip in loadbalancer.additional_vips]
        return return_vip, additional_vips

    def deallocate_vip(self, vip):
        LOG.debug("Network %s no-op, deallocate_vip vip %s",
                  self.__class__.__name__, vip.ip_address)
        self.networkconfigconfig[vip.ip_address] = (vip,
                                                    'deallocate_vip')

    def update_vip_sg(self, load_balancer, vip):
        LOG.debug("Network %s no-op, update_vip_sg loadbalancer %s, vip %s",
                  self.__class__.__name__,
                  load_balancer.id, vip.ip_address)
        self.networkconfigconfig[(load_balancer.id,
                                  vip.ip_address)] = (load_balancer, vip,
                                                      'update_vip_sg')

    def update_aap_port_sg(self,
                           load_balancer: data_models.LoadBalancer,
                           amphora: data_models.Amphora,
                           vip: data_models.Vip):
        LOG.debug("Network %s no-op, update_aap_port_sg load_balancer %s, "
                  " vip %s, amphora %s", self.__class__.__name__,
                  load_balancer.id, vip.ip_address, amphora)
        self.networkconfigconfig[(amphora.id,
                                  vip.ip_address)] = (load_balancer, vip,
                                                      amphora,
                                                      'update_aap_port_sg')

    def plug_aap_port(self, load_balancer, vip, amphora, subnet):
        LOG.debug("Network %s no-op, plug_aap_port loadbalancer %s, vip %s,"
                  " amphora %s, subnet %s",
                  self.__class__.__name__,
                  load_balancer.id, vip.ip_address, amphora, subnet)
        self.networkconfigconfig[(amphora.id,
                                  vip.ip_address)] = (
            load_balancer, vip, amphora, subnet,
            'plug_aap_port')

        fixed_ips = [network_models.FixedIP(subnet_id=subnet.id,
                                            ip_address=vip.ip_address)]

        port_id = uuidutils.generate_uuid()
        LOG.debug("Network %s no-op, plug_aap_port loadbalancer %s is using "
                  "base port ID %s",
                  self.__class__.__name__,
                  load_balancer.id, port_id)

        # Store the interface information in the no-op DB
        with self.engine.connect() as connection:
            connection.execute(insert(self.interfaces_table).values(
                port_id=port_id, network_id=vip.network_id,
                compute_id=amphora.compute_id,
                vnic_type=constants.VNIC_TYPE_NORMAL))
            for fixed_ip in fixed_ips:
                connection.execute(insert(self.fixed_ips_table).values(
                    port_id=port_id, subnet_id=fixed_ip.subnet_id,
                    ip_address=fixed_ip.ip_address))
            connection.commit()

        return data_models.Amphora(
            id=amphora.id,
            compute_id=amphora.compute_id,
            vrrp_ip='198.51.100.1',
            ha_ip='198.51.100.1',
            vrrp_port_id=port_id,
            ha_port_id=uuidutils.generate_uuid()
        )

    def unplug_vip(self, loadbalancer, vip):
        LOG.debug("Network %s no-op, unplug_vip loadbalancer %s, vip %s",
                  self.__class__.__name__,
                  loadbalancer.id, vip.ip_address)
        self.networkconfigconfig[(loadbalancer.id,
                                  vip.ip_address)] = (loadbalancer, vip,
                                                      'unplug_vip')

    def unplug_aap_port(self, vip, amphora, subnet):
        LOG.debug("Network %s no-op, unplug_aap_port vip %s amp: %s "
                  "subnet: %s",
                  self.__class__.__name__,
                  vip.ip_address, amphora.id, subnet.id)
        self.networkconfigconfig[(amphora.id,
                                  vip.ip_address)] = (vip, amphora, subnet,
                                                      'unplug_aap_port')

    def unplug_network(self, compute_id, network_id):
        LOG.debug("Network %s no-op, unplug_network compute_id %s, "
                  "network_id %s",
                  self.__class__.__name__, compute_id, network_id)
        self.networkconfigconfig[(compute_id, network_id)] = (
            compute_id, network_id, 'unplug_network')

    def get_plugged_networks(self, compute_id):
        LOG.debug("Network %s no-op, get_plugged_networks compute_id %s",
                  self.__class__.__name__, compute_id)
        self.networkconfigconfig[compute_id] = (
            compute_id, 'get_plugged_networks')

        # Retrieve the interfaces from the no-op DB for this compute_id
        interfaces = []
        with self.engine.connect() as connection:
            int_results = connection.execute(
                select(self.interfaces_table)
                .where(self.interfaces_table.c.compute_id == compute_id))
            # Walk through the matching interfaces
            for interface in int_results:
                fixed_ips = []
                # Get the fixed IPs on each interface
                fixed_ip_results = connection.execute(
                    select(self.fixed_ips_table)
                    .where(
                        self.fixed_ips_table.c.port_id == interface.port_id))
                # Build the FixedIP objects for the interface
                for fixed_ip in fixed_ip_results:
                    fixed_ips.append(network_models.FixedIP(
                        subnet_id=fixed_ip.subnet_id,
                        ip_address=fixed_ip.ip_address))
                # Add the interface object to the list
                interfaces.append(network_models.Interface(
                    compute_id=interface.compute_id,
                    network_id=interface.network_id,
                    port_id=interface.port_id, fixed_ips=fixed_ips,
                    vnic_type=interface.vnic_type))
        return interfaces

    def update_vip(self, loadbalancer, for_delete=False):
        LOG.debug("Network %s no-op, update_vip loadbalancer %s "
                  "with for delete %s",
                  self.__class__.__name__, loadbalancer, for_delete)
        self.networkconfigconfig[loadbalancer.id] = (
            loadbalancer, for_delete, 'update_vip')

    def get_network(self, network_id):
        LOG.debug("Network %s no-op, get_network network_id %s",
                  self.__class__.__name__, network_id)
        self.networkconfigconfig[network_id] = (network_id, 'get_network')
        if network_id in _NOOP_MANAGER_VARS['networks']:
            return _NOOP_MANAGER_VARS['networks'][network_id]

        network = network_models.Network(id=network_id,
                                         port_security_enabled=True)

        class ItIsInsideMe(list):
            known_subnets = None

            def __init__(self, network, parent):
                super().__init__()
                self.network = network
                self.parent = parent
                self.known_subnets = {}

            def to_dict(self, **kwargs):
                return [{}]

            def __contains__(self, item):
                self.known_subnets[item] = self.parent.get_subnet(item)
                self.known_subnets[item].network_id = self.network.id
                return True

            def __len__(self):
                return len(self.known_subnets) + 1

            def __iter__(self):
                yield from self.known_subnets
                subnet = network_models.Subnet(id=uuidutils.generate_uuid(),
                                               network_id=self.network.id)
                self.known_subnets[subnet.id] = subnet
                _NOOP_MANAGER_VARS['subnets'][subnet.id] = subnet
                yield subnet.id

        network.subnets = ItIsInsideMe(network, self)
        _NOOP_MANAGER_VARS['networks'][network_id] = network
        _NOOP_MANAGER_VARS['current_network'] = network_id
        return network

    def get_subnet(self, subnet_id):
        LOG.debug("Subnet %s no-op, get_subnet subnet_id %s",
                  self.__class__.__name__, subnet_id)
        self.networkconfigconfig[subnet_id] = (subnet_id, 'get_subnet')
        if subnet_id in _NOOP_MANAGER_VARS['subnets']:
            return _NOOP_MANAGER_VARS['subnets'][subnet_id]

        subnet = network_models.Subnet(
            id=subnet_id,
            network_id=_NOOP_MANAGER_VARS['current_network'])
        _NOOP_MANAGER_VARS['subnets'][subnet_id] = subnet
        return subnet

    def get_port(self, port_id):
        LOG.debug("Port %s no-op, get_port port_id %s",
                  self.__class__.__name__, port_id)
        self.networkconfigconfig[port_id] = (port_id, 'get_port')
        if port_id in _NOOP_MANAGER_VARS['ports']:
            return _NOOP_MANAGER_VARS['ports'][port_id]

        port = network_models.Port(id=port_id)
        _NOOP_MANAGER_VARS['ports'][port_id] = port
        return port

    def get_security_group_by_id(self, sg_id: str,
                                 context: 'context.RequestContext' = None) -> (
            'network_models.SecurityGroup'):
        LOG.debug("Network %s no-op, get_security_group_by_id %s",
                  self.__class__.__name__, sg_id)
        self.networkconfigconfig[sg_id] = (sg_id, 'get_security_group_by_id')
        return network_models.SecurityGroup(id=sg_id)

    def get_network_by_name(self, network_name):
        LOG.debug("Network %s no-op, get_network_by_name network_name %s",
                  self.__class__.__name__, network_name)
        self.networkconfigconfig[network_name] = (network_name,
                                                  'get_network_by_name')
        by_name = {n.name: n for n in _NOOP_MANAGER_VARS['networks'].values()}
        if network_name in by_name:
            return by_name[network_name]

        network = network_models.Network(id=uuidutils.generate_uuid(),
                                         port_security_enabled=True,
                                         name=network_name)
        _NOOP_MANAGER_VARS['networks'][network.id] = network
        _NOOP_MANAGER_VARS['current_network'] = network.id
        return network

    def get_subnet_by_name(self, subnet_name):
        LOG.debug("Subnet %s no-op, get_subnet_by_name subnet_name %s",
                  self.__class__.__name__, subnet_name)
        self.networkconfigconfig[subnet_name] = (subnet_name,
                                                 'get_subnet_by_name')
        by_name = {s.name: s for s in _NOOP_MANAGER_VARS['subnets'].values()}
        if subnet_name in by_name:
            return by_name[subnet_name]

        subnet = network_models.Subnet(
            id=uuidutils.generate_uuid(),
            name=subnet_name,
            network_id=_NOOP_MANAGER_VARS['current_network'])
        _NOOP_MANAGER_VARS['subnets'][subnet.id] = subnet
        return subnet

    def get_port_by_name(self, port_name):
        LOG.debug("Port %s no-op, get_port_by_name port_name %s",
                  self.__class__.__name__, port_name)
        self.networkconfigconfig[port_name] = (port_name, 'get_port_by_name')
        by_name = {p.name: p for p in _NOOP_MANAGER_VARS['ports'].values()}
        if port_name in by_name:
            return by_name[port_name]

        port = network_models.Port(id=uuidutils.generate_uuid(),
                                   name=port_name)
        _NOOP_MANAGER_VARS['ports'][port.id] = port
        return port

    def get_port_by_net_id_device_id(self, network_id, device_id):
        LOG.debug("Port %s no-op, get_port_by_net_id_device_id network_id %s"
                  " device_id %s",
                  self.__class__.__name__, network_id, device_id)
        self.networkconfigconfig[(network_id, device_id)] = (
            network_id, device_id, 'get_port_by_net_id_device_id')
        by_net_dev_id = {(p.network_id, p.device_id): p
                         for p in _NOOP_MANAGER_VARS['ports'].values()}
        if (network_id, device_id) in by_net_dev_id:
            return by_net_dev_id[(network_id, device_id)]

        port = network_models.Port(id=uuidutils.generate_uuid(),
                                   network_id=network_id,
                                   device_id=device_id)
        _NOOP_MANAGER_VARS['ports'][port.id] = port
        return port

    def get_security_group(self, sg_name):
        LOG.debug("Network %s no-op, get_security_group name %s",
                  self.__class__.__name__, sg_name)
        self.networkconfigconfig[(sg_name)] = (sg_name, 'get_security_group')
        return network_models.SecurityGroup(id=uuidutils.generate_uuid())

    def failover_preparation(self, amphora):
        LOG.debug("failover %s no-op, failover_preparation, amphora id %s",
                  self.__class__.__name__, amphora.id)

    def plug_port(self, amphora, port):
        LOG.debug("Network %s no-op, plug_port amphora.id %s, port_id "
                  "%s", self.__class__.__name__, amphora.id, port.id)
        self.networkconfigconfig[(amphora.id, port.id)] = (
            amphora, port, 'plug_port')

    def _get_amp_net_configs(self, amp, amp_configs, vip_subnet, vip_port):
        vrrp_port = self.get_port(amp.vrrp_port_id)
        ha_port = self.get_port(amp.ha_port_id)
        amp_configs[amp.id] = network_models.AmphoraNetworkConfig(
            amphora=amp,
            vrrp_subnet=self.get_subnet(
                vrrp_port.get_subnet_id(amp.vrrp_ip)),
            vrrp_port=vrrp_port,
            ha_subnet=self.get_subnet(
                ha_port.get_subnet_id(amp.ha_ip)),
            ha_port=ha_port)

    def get_network_configs(self, loadbalancer, amphora=None):
        amphora_id = amphora.id if amphora else None
        LOG.debug("Network %s no-op, get_network_configs loadbalancer id "
                  "%s amphora id: %s", self.__class__.__name__,
                  loadbalancer.id, amphora_id)
        self.networkconfigconfig[(loadbalancer.id)] = (
            loadbalancer, 'get_network_configs')
        vip_subnet = self.get_subnet(loadbalancer.vip.subnet_id)
        vip_port = self.get_port(loadbalancer.vip.port_id)

        amp_configs = {}
        if amphora:
            self._get_amp_net_configs(amphora, amp_configs,
                                      vip_subnet, vip_port)
        else:
            for amp in loadbalancer.amphorae:
                self._get_amp_net_configs(amp, amp_configs,
                                          vip_subnet, vip_port)

        return amp_configs

    def get_qos_policy(self, qos_policy_id):
        LOG.debug("Qos Policy %s no-op, get_qos_policy qos_policy_id %s",
                  self.__class__.__name__, qos_policy_id)
        self.networkconfigconfig[qos_policy_id] = (qos_policy_id,
                                                   'get_qos_policy')
        return qos_policy_id

    def apply_qos_on_port(self, qos_id, port_id):
        LOG.debug("Network %s no-op, apply_qos_on_port qos_id %s, port_id "
                  "%s", self.__class__.__name__, qos_id, port_id)
        self.networkconfigconfig[(qos_id, port_id)] = (
            qos_id, port_id, 'apply_qos_on_port')

    def qos_enabled(self):
        return self._qos_extension_enabled

    def get_network_ip_availability(self, network):
        LOG.debug("Network %s no-op, network_ip_availability network_id %s",
                  self.__class__.__name__, network.id)
        self.networkconfigconfig[(network.id, 'ip_availability')] = (
            network.id, 'get_network_ip_availability')
        ip_avail = network_models.Network_IP_Availability(
            network_id=network.id)
        subnet_ip_availability = []
        for subnet_id in list(network.subnets):
            subnet_ip_availability.append({'subnet_id': subnet_id,
                                          'used_ips': 0, 'total_ips': 254})
        ip_avail.subnet_ip_availability = subnet_ip_availability
        return ip_avail

    def delete_port(self, port_id):
        LOG.debug("Network %s no-op, delete_port port_id %s",
                  self.__class__.__name__, port_id)
        self.networkconfigconfig[port_id] = (port_id, 'delete_port')

        # Remove the port from the no-op DB
        with self.engine.connect() as connection:
            connection.execute(delete(self.interfaces_table)
                               .where(
                                   self.interfaces_table.c.port_id == port_id))
            connection.execute(delete(self.fixed_ips_table)
                               .where(
                                   self.fixed_ips_table.c.port_id == port_id))
            connection.commit()

    def set_port_admin_state_up(self, port_id, state):
        LOG.debug("Network %s no-op, set_port_admin_state_up port_id %s, "
                  "state %s", self.__class__.__name__, port_id, state)
        self.networkconfigconfig[(port_id, state)] = (port_id, state,
                                                      'admin_down_port')

    def create_port(self, network_id, name=None, fixed_ips=(),
                    secondary_ips=(), security_group_ids=(),
                    admin_state_up=True, qos_policy_id=None,
                    vnic_type=constants.VNIC_TYPE_NORMAL):
        LOG.debug("Network %s no-op, create_port network_id %s",
                  self.__class__.__name__, network_id)
        if not name:
            name = 'no-op-port'
        port_id = uuidutils.generate_uuid()
        project_id = uuidutils.generate_uuid()

        fixed_ip_obj_list = []
        for fixed_ip in fixed_ips:
            if fixed_ip and not fixed_ip.get('ip_address'):
                fixed_ip_obj_list.append(
                    network_models.FixedIP(subnet_id=fixed_ip.get('subnet_id'),
                                           ip_address='198.51.100.56'))
            else:
                fixed_ip_obj_list.append(
                    network_models.FixedIP(
                        subnet_id=fixed_ip.get('subnet_id'),
                        ip_address=fixed_ip.get('ip_address')))
        if not fixed_ip_obj_list:
            fixed_ip_obj_list = [network_models.FixedIP(
                subnet_id=uuidutils.generate_uuid(),
                ip_address='198.51.100.56')]

        self.networkconfigconfig[(network_id, 'create_port')] = (
            network_id, name, fixed_ip_obj_list, secondary_ips,
            security_group_ids, admin_state_up, qos_policy_id, vnic_type)

        # Store the interface information in the no-op DB
        with self.engine.connect() as connection:
            connection.execute(insert(self.interfaces_table).values(
                port_id=port_id, network_id=network_id, vnic_type=vnic_type))
            for fixed_ip in fixed_ip_obj_list:
                connection.execute(insert(self.fixed_ips_table).values(
                    port_id=port_id, subnet_id=fixed_ip.subnet_id,
                    ip_address=fixed_ip.ip_address))
            connection.commit()

        return network_models.Port(
            id=port_id, name=name, device_id='no-op-device-id',
            device_owner='Octavia', mac_address='00:00:5E:00:53:05',
            network_id=network_id, status='UP', project_id=project_id,
            admin_state_up=admin_state_up, fixed_ips=fixed_ip_obj_list,
            qos_policy_id=qos_policy_id, security_group_ids=security_group_ids,
            vnic_type=vnic_type)

    def plug_fixed_ip(self, port_id, subnet_id, ip_address=None):
        LOG.debug("Network %s no-op, plug_fixed_ip port_id %s, subnet_id "
                  "%s, ip_address %s", self.__class__.__name__, port_id,
                  subnet_id, ip_address)
        self.networkconfigconfig[(port_id, subnet_id)] = (
            port_id, subnet_id, ip_address, 'plug_fixed_ip')

        port = network_models.Port(id=port_id,
                                   network_id=uuidutils.generate_uuid())
        _NOOP_MANAGER_VARS['ports'][port.id] = port
        return port

    def unplug_fixed_ip(self, port_id, subnet_id):
        LOG.debug("Network %s no-op, unplug_fixed_ip port_id %s, subnet_id "
                  "%s", self.__class__.__name__, port_id,
                  subnet_id)
        self.networkconfigconfig[(port_id, subnet_id)] = (
            port_id, subnet_id, 'unplug_fixed_ip')

        return _NOOP_MANAGER_VARS['ports'].pop(port_id, None)


class NoopNetworkDriver(driver_base.AbstractNetworkDriver):
    def __init__(self):
        super().__init__()
        self.driver = NoopManager()

    def allocate_vip(self, loadbalancer):
        return self.driver.allocate_vip(loadbalancer)

    def deallocate_vip(self, vip):
        self.driver.deallocate_vip(vip)

    def unplug_vip(self, loadbalancer, vip):
        self.driver.unplug_vip(loadbalancer, vip)

    def unplug_network(self, compute_id, network_id):
        self.driver.unplug_network(compute_id, network_id)

    def get_plugged_networks(self, compute_id):
        return self.driver.get_plugged_networks(compute_id)

    def update_vip(self, loadbalancer, for_delete=False):
        self.driver.update_vip(loadbalancer, for_delete)

    def get_network(self, network_id, context=None):
        return self.driver.get_network(network_id)

    def get_subnet(self, subnet_id, context=None):
        return self.driver.get_subnet(subnet_id)

    def get_port(self, port_id, context=None):
        return self.driver.get_port(port_id)

    def get_security_group_by_id(self, sg_id: str,
                                 context: 'context.RequestContext' = None) -> (
            'network_models.SecurityGroup'):
        return self.driver.get_security_group_by_id(sg_id, context=context)

    def get_qos_policy(self, qos_policy_id):
        return self.driver.get_qos_policy(qos_policy_id)

    def get_network_by_name(self, network_name):
        return self.driver.get_network_by_name(network_name)

    def get_subnet_by_name(self, subnet_name):
        return self.driver.get_subnet_by_name(subnet_name)

    def get_port_by_name(self, port_name):
        return self.driver.get_port_by_name(port_name)

    def get_port_by_net_id_device_id(self, network_id, device_id):
        return self.driver.get_port_by_net_id_device_id(network_id, device_id)

    def get_security_group(self, sg_name):
        return self.driver.get_security_group(sg_name)

    def failover_preparation(self, amphora):
        self.driver.failover_preparation(amphora)

    def plug_port(self, amphora, port):
        return self.driver.plug_port(amphora, port)

    def get_network_configs(self, loadbalancer, amphora=None):
        return self.driver.get_network_configs(loadbalancer, amphora)

    def apply_qos_on_port(self, qos_id, port_id):
        self.driver.apply_qos_on_port(qos_id, port_id)

    def update_vip_sg(self, load_balancer, vip):
        self.driver.update_vip_sg(load_balancer, vip)

    def update_aap_port_sg(self,
                           load_balancer: data_models.LoadBalancer,
                           amphora: data_models.Amphora,
                           vip: data_models.Vip):
        self.driver.update_aap_port_sg(load_balancer, amphora, vip)

    def plug_aap_port(self, load_balancer, vip, amphora, subnet):
        return self.driver.plug_aap_port(load_balancer, vip, amphora, subnet)

    def unplug_aap_port(self, vip, amphora, subnet):
        self.driver.unplug_aap_port(vip, amphora, subnet)

    def qos_enabled(self):
        return self.driver.qos_enabled()

    def get_network_ip_availability(self, network):
        return self.driver.get_network_ip_availability(network)

    def delete_port(self, port_id):
        self.driver.delete_port(port_id)

    def set_port_admin_state_up(self, port_id, state):
        self.driver.set_port_admin_state_up(port_id, state)

    def create_port(self, network_id, name=None, fixed_ips=(),
                    secondary_ips=(), security_group_ids=(),
                    admin_state_up=True, qos_policy_id=None,
                    vnic_type=constants.VNIC_TYPE_NORMAL):
        return self.driver.create_port(
            network_id, name, fixed_ips, secondary_ips, security_group_ids,
            admin_state_up, qos_policy_id, vnic_type)

    def plug_fixed_ip(self, port_id, subnet_id, ip_address=None):
        return self.driver.plug_fixed_ip(port_id, subnet_id, ip_address)

    def unplug_fixed_ip(self, port_id, subnet_id):
        return self.driver.unplug_fixed_ip(port_id, subnet_id)



# Copyright 2020 Red Hat, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import ipaddress
import json
import os
import stat

from oslo_config import cfg

from octavia.common import constants as consts

CONF = cfg.CONF


class InterfaceFile:
    def __init__(self, name, if_type, mtu=None, addresses=None,
                 routes=None, rules=None, scripts=None, is_sriov=False):
        self.name = name
        self.if_type = if_type
        self.mtu = mtu
        self.addresses = addresses or []
        self.routes = routes or []
        self.rules = rules or []
        self.scripts = scripts or {
            consts.IFACE_UP: [],
            consts.IFACE_DOWN: []
        }
        self.is_sriov = is_sriov

    @classmethod
    def get_extensions(cls):
        return [".json"]

    @classmethod
    def load(cls, fp):
        return json.load(fp)

    @classmethod
    def dump(cls, obj):
        return json.dumps(obj)

    @classmethod
    def from_file(cls, filename):
        with open(filename, encoding='utf-8') as fp:
            config = cls.load(fp)

        return InterfaceFile(**config)

    @classmethod
    def get_directory(cls):
        return (CONF.amphora_agent.agent_server_network_dir or
                consts.AMP_NET_DIR_TEMPLATE)

    @classmethod
    def get_host_routes(cls, routes, **kwargs):
        host_routes = []
        if routes:
            for hr in routes:
                route = {
                    consts.DST: hr['destination'],
                    consts.GATEWAY: hr['nexthop'],
                    consts.FLAGS: [consts.ONLINK]
                }
                route.update(kwargs)
                host_routes.append(route)
        return host_routes

    def write(self):
        mode = stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IROTH
        flags = os.O_WRONLY | os.O_CREAT | os.O_TRUNC

        net_dir = self.get_directory()

        try:
            os.makedirs(net_dir)
        except OSError:
            pass

        interface_file = f"{self.name}.json"

        with os.fdopen(os.open(os.path.join(net_dir, interface_file),
                               flags, mode), 'w') as fp:
            interface = {
                consts.NAME: self.name,
                consts.IF_TYPE: self.if_type,
                consts.ADDRESSES: self.addresses,
                consts.ROUTES: self.routes,
                consts.RULES: self.rules,
                consts.SCRIPTS: self.scripts,
                consts.IS_SRIOV: self.is_sriov
            }
            if self.mtu:
                interface[consts.MTU] = self.mtu
            fp.write(self.dump(interface))


class VIPInterfaceFile(InterfaceFile):
    def __init__(self, name, mtu, vips, vrrp_info, fixed_ips, topology,
                 is_sriov=False):

        super().__init__(name, if_type=consts.VIP, mtu=mtu, is_sriov=is_sriov)

        has_ipv4 = any(vip['ip_version'] == 4 for vip in vips)
        has_ipv6 = any(vip['ip_version'] == 6 for vip in vips)

        if vrrp_info:
            self.addresses.append({
                consts.ADDRESS: vrrp_info['ip'],
                consts.PREFIXLEN: vrrp_info['prefixlen']
            })
        else:
            if has_ipv4:
                self.addresses.append({
                    consts.DHCP: True
                })
            if has_ipv6:
                self.addresses.append({
                    consts.IPV6AUTO: True
                })

        ip_versions = set()

        for vip in vips:
            gateway = vip.get('gateway')
            ip_version = vip['ip_version']
            ip_versions.add(ip_version)

            if gateway:
                # Add default routes if there's a gateway
                self.routes.append({
                    consts.DST: (
                        "::/0" if ip_version == 6 else "0.0.0.0/0"),
                    consts.GATEWAY: gateway,
                    consts.FLAGS: [consts.ONLINK]
                })
                if topology != consts.TOPOLOGY_ACTIVE_STANDBY:
                    self.routes.append({
                        consts.DST: (
                            "::/0" if ip_version == 6 else "0.0.0.0/0"),
                        consts.GATEWAY: gateway,
                        consts.FLAGS: [consts.ONLINK],
                        consts.TABLE: 1,
                    })

            # In ACTIVE_STANDBY topology, keepalived configures the VIP
            # address. Keep track of it in the interface file but mark it with
            # a special flag so the amphora-interface would not add/delete
            # keepalived-maintained things.
            self.addresses.append({
                consts.ADDRESS: vip['ip_address'],
                consts.PREFIXLEN: 128 if ip_version == 6 else 32,
                # OCTAVIA_OWNED = False when this address is managed by another
                # tool (keepalived)
                consts.OCTAVIA_OWNED: (
                    topology != consts.TOPOLOGY_ACTIVE_STANDBY)
            })

            vip_cidr = ipaddress.ip_network(
                f"{vip['ip_address']}/{vip['prefixlen']}",
                strict=False)
            self.routes.append({
                consts.DST: vip_cidr.exploded,
                consts.SCOPE: 'link'
            })

            if topology != consts.TOPOLOGY_ACTIVE_STANDBY:
                self.routes.append({
                    consts.DST: vip_cidr.exploded,
                    consts.PREFSRC: vip['ip_address'],
                    consts.SCOPE: 'link',
                    consts.TABLE: 1
                })
                self.rules.append({
                    consts.SRC: vip['ip_address'],
                    consts.SRC_LEN: 128 if ip_version == 6 else 32,
                    consts.TABLE: 1
                })

            self.routes.extend(self.get_host_routes(vip['host_routes']))
            self.routes.extend(self.get_host_routes(vip['host_routes'],
                                                    table=1))

        for fixed_ip in fixed_ips or ():
            ip_addr = fixed_ip['ip_address']
            cidr = fixed_ip['subnet_cidr']
            ip = ipaddress.ip_address(ip_addr)
            network = ipaddress.ip_network(cidr)
            prefixlen = network.prefixlen
            self.addresses.append({
                consts.ADDRESS: fixed_ip['ip_address'],
                consts.PREFIXLEN: prefixlen,
            })

            ip_versions.add(ip.version)

            gateway = fixed_ip.get('gateway')
            if gateway:
                # Add default routes if there's a gateway
                self.routes.append({
                    consts.DST: (
                        "::/0" if ip.version == 6 else "0.0.0.0/0"),
                    consts.GATEWAY: gateway,
                    consts.FLAGS: [consts.ONLINK]
                })
                if topology != consts.TOPOLOGY_ACTIVE_STANDBY:
                    self.routes.append({
                        consts.DST: (
                            "::/0" if ip.version == 6 else "0.0.0.0/0"),
                        consts.GATEWAY: gateway,
                        consts.FLAGS: [consts.ONLINK],
                        consts.TABLE: 1,
                    })

            host_routes = self.get_host_routes(
                fixed_ip.get('host_routes', []))
            self.routes.extend(host_routes)

        if is_sriov:
            sriov_param = ' sriov'
        else:
            sriov_param = ''

        for ip_v in ip_versions:
            self.scripts[consts.IFACE_UP].append({
                consts.COMMAND: (
                    "/usr/local/bin/lvs-masquerade.sh add {} {}{}".format(
                        'ipv6' if ip_v == 6 else 'ipv4', name, sriov_param))
            })
            self.scripts[consts.IFACE_DOWN].append({
                consts.COMMAND: (
                    "/usr/local/bin/lvs-masquerade.sh delete {} {}{}".format(
                        'ipv6' if ip_v == 6 else 'ipv4', name, sriov_param))
            })


class PortInterfaceFile(InterfaceFile):
    def __init__(self, name, mtu, fixed_ips, is_sriov=False):
        super().__init__(name, if_type=consts.BACKEND, mtu=mtu,
                         is_sriov=is_sriov)

        if fixed_ips:
            ip_versions = set()

            for fixed_ip in fixed_ips:
                ip_addr = fixed_ip['ip_address']
                cidr = fixed_ip['subnet_cidr']
                ip = ipaddress.ip_address(ip_addr)
                network = ipaddress.ip_network(cidr)
                prefixlen = network.prefixlen
                self.addresses.append({
                    consts.ADDRESS: fixed_ip['ip_address'],
                    consts.PREFIXLEN: prefixlen,
                })

                ip_versions.add(ip.version)

                host_routes = self.get_host_routes(
                    fixed_ip.get('host_routes', []))
                self.routes.extend(host_routes)
        else:
            ip_versions = {4, 6}

            self.addresses.append({
                consts.DHCP: True,
                consts.IPV6AUTO: True
            })

        if is_sriov:
            sriov_param = ' sriov'
        else:
            sriov_param = ''

        for ip_version in ip_versions:
            self.scripts[consts.IFACE_UP].append({
                consts.COMMAND: (
                    "/usr/local/bin/lvs-masquerade.sh add {} {}{}".format(
                        'ipv6' if ip_version == 6 else 'ipv4', name,
                        sriov_param))
            })
            self.scripts[consts.IFACE_DOWN].append({
                consts.COMMAND: (
                    "/usr/local/bin/lvs-masquerade.sh delete {} {}{}".format(
                        'ipv6' if ip_version == 6 else 'ipv4', name,
                        sriov_param))
            })



# Copyright 2015 Hewlett Packard Enterprise Development Company LP
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
import copy
from unittest import mock

from oslo_config import cfg
from oslo_config import fixture as oslo_fixture

from octavia.amphorae.drivers.keepalived.jinja import jinja_cfg
from octavia.common import constants
from octavia.network import data_models as n_data_models
import octavia.tests.unit.base as base


class TestVRRPRestDriver(base.TestCase):

    def setUp(self):
        super().setUp()
        conf = self.useFixture(oslo_fixture.Config(cfg.CONF))
        conf.config(group="haproxy_amphora", base_path='/tmp/test')
        conf.config(group="keepalived_vrrp", vrrp_garp_refresh_interval=5)
        conf.config(group="keepalived_vrrp", vrrp_garp_refresh_count=2)
        conf.config(group="keepalived_vrrp", vrrp_check_interval=5)
        conf.config(group="keepalived_vrrp", vrrp_fail_count=2)
        conf.config(group="keepalived_vrrp", vrrp_success_count=2)

        self.templater = jinja_cfg.KeepalivedJinjaTemplater()

        self.amphora1 = mock.MagicMock()
        self.amphora1.status = constants.AMPHORA_ALLOCATED
        self.amphora1.vrrp_ip = '10.0.0.1'
        self.amphora1.role = constants.ROLE_MASTER
        self.amphora1.vrrp_interface = 'eth1'
        self.amphora1.vrrp_id = 1
        self.amphora1.vrrp_priority = 100

        self.amphora2 = mock.MagicMock()
        self.amphora2.status = constants.AMPHORA_ALLOCATED
        self.amphora2.vrrp_ip = '10.0.0.2'
        self.amphora2.role = constants.ROLE_BACKUP
        self.amphora2.vrrp_interface = 'eth1'
        self.amphora2.vrrp_id = 1
        self.amphora2.vrrp_priority = 90

        self.lb = mock.MagicMock()
        self.lb.amphorae = [self.amphora1, self.amphora2]
        self.lb.vrrp_group.vrrp_group_name = 'TESTGROUP'
        self.lb.vrrp_group.vrrp_auth_type = constants.VRRP_AUTH_DEFAULT
        self.lb.vrrp_group.vrrp_auth_pass = 'TESTPASSWORD'
        self.lb.vip.ip_address = '10.1.0.5'
        self.lb.vrrp_group.advert_int = 10

        self.ref_conf = (
            "vrrp_script check_script {\n"
            "    script /tmp/test/vrrp/check_script.sh\n"
            "    interval 5\n"
            "    fall 2\n"
            "    rise 2\n"
            "}\n"
            "\n"
            "vrrp_instance TESTGROUP {\n"
            "    interface eth1\n"
            "    virtual_router_id 1\n"
            "    priority 100\n"
            "    nopreempt\n"
            "    accept\n"
            "    garp_master_refresh 5\n"
            "    garp_master_refresh_repeat 2\n"
            "    advert_int 10\n"
            "    authentication {\n"
            "        auth_type PASS\n"
            "        auth_pass TESTPASSWORD\n"
            "    }\n"
            "\n"
            "    unicast_src_ip 10.0.0.1\n"
            "    unicast_peer {\n"
            "        10.0.0.2\n"
            "    }\n"
            "\n"
            "    virtual_ipaddress {\n"
            "        10.1.0.5\n"
            "    }\n\n"
            "    virtual_ipaddress_excluded {\n"
            "    }\n\n"
            "    virtual_routes {\n"
            "        10.1.0.0/24 dev eth1 src 10.1.0.5 scope link table 1\n"
            "        default via 10.1.0.1 dev eth1 onlink table 1\n"
            "    }\n\n"
            "    virtual_rules {\n"
            "        from 10.1.0.5/32 table 1 priority 100\n"
            "    }\n\n"
            "    track_script {\n"
            "        check_script\n"
            "    }\n"
            "}")

        self.amphora1v6 = copy.deepcopy(self.amphora1)
        self.amphora1v6.vrrp_ip = '2001:db8::10'
        self.amphora2v6 = copy.deepcopy(self.amphora2)
        self.amphora2v6.vrrp_ip = '2001:db8::11'
        self.lbv6 = copy.deepcopy(self.lb)
        self.lbv6.amphorae = [self.amphora1v6, self.amphora2v6]
        self.lbv6.vip.ip_address = '2001:db8::15'

        self.ref_v6_conf = (
            "vrrp_script check_script {\n"
            "    script /tmp/test/vrrp/check_script.sh\n"
            "    interval 5\n"
            "    fall 2\n"
            "    rise 2\n"
            "}\n"
            "\n"
            "vrrp_instance TESTGROUP {\n"
            "    interface eth1\n"
            "    virtual_router_id 1\n"
            "    priority 100\n"
            "    nopreempt\n"
            "    accept\n"
            "    garp_master_refresh 5\n"
            "    garp_master_refresh_repeat 2\n"
            "    advert_int 10\n"
            "    authentication {\n"
            "        auth_type PASS\n"
            "        auth_pass TESTPASSWORD\n"
            "    }\n"
            "\n"
            "    unicast_src_ip 2001:db8::10\n"
            "    unicast_peer {\n"
            "        2001:db8::11\n"
            "    }\n"
            "\n"
            "    virtual_ipaddress {\n"
            "        2001:db8::15\n"
            "    }\n\n"
            "    virtual_ipaddress_excluded {\n"
            "    }\n\n"
            "    virtual_routes {\n"
            "        2001:db8::/64 dev eth1 src "
            "2001:db8::15 scope link table 1\n"
            "        default via 2001:db8::ff dev eth1 onlink table 1\n"
            "    }\n\n"
            "    virtual_rules {\n"
            "        from 2001:db8::15/128 table 1 priority 100\n"
            "    }\n\n"
            "    track_script {\n"
            "        check_script\n"
            "    }\n"
            "}")

        self.ref_v4_v6_conf = (
            "vrrp_script check_script {\n"
            "    script /tmp/test/vrrp/check_script.sh\n"
            "    interval 5\n"
            "    fall 2\n"
            "    rise 2\n"
            "}\n"
            "\n"
            "vrrp_instance TESTGROUP {\n"
            "    interface eth1\n"
            "    virtual_router_id 1\n"
            "    priority 100\n"
            "    nopreempt\n"
            "    accept\n"
            "    garp_master_refresh 5\n"
            "    garp_master_refresh_repeat 2\n"
            "    advert_int 10\n"
            "    authentication {\n"
            "        auth_type PASS\n"
            "        auth_pass TESTPASSWORD\n"
            "    }\n"
            "\n"
            "    unicast_src_ip 10.0.0.1\n"
            "    unicast_peer {\n"
            "        10.0.0.2\n"
            "    }\n"
            "\n"
            "    virtual_ipaddress {\n"
            "        10.1.0.5\n"
            "    }\n\n"
            "    virtual_ipaddress_excluded {\n"
            "        2001:db8::15\n"
            "    }\n\n"
            "    virtual_routes {\n"
            "        10.1.0.0/24 dev eth1 src 10.1.0.5 scope link table 1\n"
            "        default via 10.1.0.1 dev eth1 onlink table 1\n"
            "        2001:db8::/64 dev eth1 src "
            "2001:db8::15 scope link table 1\n"
            "        default via 2001:db8::ff dev eth1 onlink table 1\n"
            "    }\n\n"
            "    virtual_rules {\n"
            "        from 10.1.0.5/32 table 1 priority 100\n"
            "        from 2001:db8::15/128 table 1 priority 100\n"
            "    }\n\n"
            "    track_script {\n"
            "        check_script\n"
            "    }\n"
            "}")

        self.ref_v6_v4_conf = (
            "vrrp_script check_script {\n"
            "    script /tmp/test/vrrp/check_script.sh\n"
            "    interval 5\n"
            "    fall 2\n"
            "    rise 2\n"
            "}\n"
            "\n"
            "vrrp_instance TESTGROUP {\n"
            "    interface eth1\n"
            "    virtual_router_id 1\n"
            "    priority 100\n"
            "    nopreempt\n"
            "    accept\n"
            "    garp_master_refresh 5\n"
            "    garp_master_refresh_repeat 2\n"
            "    advert_int 10\n"
            "    authentication {\n"
            "        auth_type PASS\n"
            "        auth_pass TESTPASSWORD\n"
            "    }\n"
            "\n"
            "    unicast_src_ip 2001:db8::10\n"
            "    unicast_peer {\n"
            "        2001:db8::11\n"
            "    }\n"
            "\n"
            "    virtual_ipaddress {\n"
            "        2001:db8::15\n"
            "    }\n\n"
            "    virtual_ipaddress_excluded {\n"
            "        10.1.0.5\n"
            "    }\n\n"
            "    virtual_routes {\n"
            "        2001:db8::/64 dev eth1 src "
            "2001:db8::15 scope link table 1\n"
            "        default via 2001:db8::ff dev eth1 onlink table 1\n"
            "        10.1.0.0/24 dev eth1 src 10.1.0.5 scope link table 1\n"
            "        default via 10.1.0.1 dev eth1 onlink table 1\n"
            "    }\n\n"
            "    virtual_rules {\n"
            "        from 2001:db8::15/128 table 1 priority 100\n"
            "        from 10.1.0.5/32 table 1 priority 100\n"
            "    }\n\n"
            "    track_script {\n"
            "        check_script\n"
            "    }\n"
            "}")

    def test_build_keepalived_config(self):
        mock_subnet = n_data_models.Subnet()
        mock_subnet.cidr = '10.1.0.0/24'
        mock_subnet.gateway_ip = '10.1.0.1'
        mock_subnet.host_routes = []
        amp_net_config = n_data_models.AmphoraNetworkConfig(
            vip_subnet=mock_subnet).to_dict(recurse=True)

        config = self.templater.build_keepalived_config(
            self.lb, self.amphora1, amp_net_config)
        self.assertEqual(self.ref_conf, config)

    def test_build_keepalived_ipv6_config(self):
        mock_subnet = n_data_models.Subnet()
        mock_subnet.cidr = '2001:db8::/64'
        mock_subnet.gateway_ip = '2001:db8::ff'
        mock_subnet.host_routes = []
        amp_net_config = n_data_models.AmphoraNetworkConfig(
            vip_subnet=mock_subnet).to_dict(recurse=True)

        config = self.templater.build_keepalived_config(
            self.lbv6, self.amphora1v6, amp_net_config)
        self.assertEqual(self.ref_v6_conf, config)

    def test_build_keepalived_config_with_additional_vips(self):
        mock_subnet1 = n_data_models.Subnet()
        mock_subnet1.cidr = '10.1.0.0/24'
        mock_subnet1.gateway_ip = '10.1.0.1'
        mock_subnet1.host_routes = []
        mock_subnet2 = n_data_models.Subnet()
        mock_subnet2.cidr = '2001:db8::/64'
        mock_subnet2.gateway_ip = '2001:db8::ff'
        mock_subnet2.host_routes = []

        # Use IPv4 as the primary VIP, IPv6 as secondary
        additional_vip = n_data_models.AdditionalVipData(
            ip_address=self.lbv6.vip.ip_address,
            subnet=mock_subnet2
        )
        amp_net_config = n_data_models.AmphoraNetworkConfig(
            vip_subnet=mock_subnet1,
            additional_vip_data=[additional_vip]).to_dict(recurse=True)

        config = self.templater.build_keepalived_config(
            self.lb, self.amphora1, amp_net_config)
        self.assertEqual(self.ref_v4_v6_conf, config)

        # Use IPv6 as the primary VIP, IPv4 as secondary
        additional_vip = n_data_models.AdditionalVipData(
            ip_address=self.lb.vip.ip_address,
            subnet=mock_subnet1
        )
        amp_net_config = n_data_models.AmphoraNetworkConfig(
            vip_subnet=mock_subnet2,
            additional_vip_data=[additional_vip]).to_dict(recurse=True)

        config = self.templater.build_keepalived_config(
            self.lbv6, self.amphora1v6, amp_net_config)
        self.assertEqual(self.ref_v6_v4_conf, config)

    def test_build_keepalived_config_with_additional_vips_v2(self):
        subnet1 = {
            "cidr": '10.1.0.0/24',
            "gateway_ip": '10.1.0.1',
            "host_routes": []
        }
        subnet2 = {
            "cidr": '2001:db8::/64',
            "gateway_ip": '2001:db8::ff',
            "host_routes": []
        }

        # Use IPv4 as the primary VIP, IPv6 as secondary
        additional_vip = {
            "ip_address": self.lbv6.vip.ip_address,
            "subnet": subnet2
        }
        amp_net_config = {
            "vip_subnet": subnet1,
            "additional_vip_data": [additional_vip]
        }

        config = self.templater.build_keepalived_config(
            self.lb, self.amphora1, amp_net_config)
        self.assertEqual(self.ref_v4_v6_conf, config)

        # Use IPv6 as the primary VIP, IPv4 as secondary
        additional_vip = {
            "ip_address": self.lb.vip.ip_address,
            "subnet": subnet1
        }
        amp_net_config = {
            "vip_subnet": subnet2,
            "additional_vip_data": [additional_vip]
        }

        config = self.templater.build_keepalived_config(
            self.lbv6, self.amphora1v6, amp_net_config)
        self.assertEqual(self.ref_v6_v4_conf, config)



# Copyright 2011, VMware, Inc., 2014 A10 Networks
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# Borrowed from nova code base, more utilities will be added/borrowed as and
# when needed.

"""Utilities and helper functions."""

import base64
import hashlib
import ipaddress
import re
import socket
import typing

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils
from stevedore import driver as stevedore_driver

from octavia.common import constants
if typing.TYPE_CHECKING:
    from octavia.network import base as network_base

CONF = cfg.CONF

LOG = logging.getLogger(__name__)


def get_hostname():
    return socket.gethostname()


def base64_sha1_string(string_to_hash):
    """Get a b64-encoded sha1 hash of a string. Not intended to be secure!"""
    # TODO(rm_work): applying nosec here because this is not intended to be
    # secure, it's just a way to get a consistent ID. Changing this would
    # break backwards compatibility with existing loadbalancers.
    hash_str = hashlib.sha1(string_to_hash.encode('utf-8')).digest()  # nosec
    b64_str = base64.b64encode(hash_str, str.encode('_-', 'ascii'))
    b64_sha1 = b64_str.decode('UTF-8')
    # https://github.com/haproxy/haproxy/issues/644
    return re.sub(r"^-", "x", b64_sha1)


def get_amphora_driver():
    amphora_driver = stevedore_driver.DriverManager(
        namespace='octavia.amphora.drivers',
        name=CONF.controller_worker.amphora_driver,
        invoke_on_load=True
    ).driver
    return amphora_driver


def get_network_driver() -> 'network_base.AbstractNetworkDriver':
    CONF.import_group('controller_worker', 'octavia.common.config')
    network_driver = stevedore_driver.DriverManager(
        namespace='octavia.network.drivers',
        name=CONF.controller_worker.network_driver,
        invoke_on_load=True
    ).driver
    return network_driver


def ip_version(ip_address):
    ip = ipaddress.ip_address(ip_address)
    return ip.version


def is_ipv4(ip_address):
    """Check if ip address is IPv4 address."""
    ip = ipaddress.ip_address(ip_address)
    return ip.version == 4


def is_ipv6(ip_address):
    """Check if ip address is IPv6 address."""
    ip = ipaddress.ip_address(ip_address)
    return ip.version == 6


def is_cidr_ipv6(cidr):
    """Check if CIDR is IPv6 address with subnet prefix."""
    ip = ipaddress.ip_network(cidr, strict=False)
    return ip.version == 6


def is_ipv6_lla(ip_address):
    """Check if ip address is IPv6 link local address."""
    ip = ipaddress.ip_address(ip_address)
    return ip.version == 6 and ip.is_link_local


def ip_port_str(ip_address, port):
    """Return IP port as string representation depending on address family."""
    ip = ipaddress.ip_address(ip_address)
    if ip.version == 4:
        return f"{ip}:{port}"
    return f"[{ip}]:{port}"


def netmask_to_prefix(netmask):
    return ipaddress.ip_network(f'0.0.0.0/{netmask}', strict=False).prefixlen


def ip_netmask_to_cidr(ip, netmask):
    return ipaddress.ip_network(f'{ip}/{netmask}', strict=False).with_prefixlen


def get_vip_security_group_name(loadbalancer_id):
    if loadbalancer_id:
        return constants.VIP_SECURITY_GROUP_PREFIX + loadbalancer_id
    return None


def get_compatible_value(value):
    if isinstance(value, str):
        value = value.encode('utf-8')
    return value


def get_compatible_server_certs_key_passphrase():
    key = CONF.certificates.server_certs_key_passphrase
    if isinstance(key, str):
        key = key.encode('utf-8')
    return base64.urlsafe_b64encode(key)


def subnet_ip_availability(nw_ip_avail, subnet_id, req_num_ips):
    for subnet in nw_ip_avail.subnet_ip_availability:
        if subnet['subnet_id'] == subnet_id:
            return subnet['total_ips'] - subnet['used_ips'] >= req_num_ips
    return None


def b(s):
    return s.encode('utf-8')


def expand_expected_codes(codes):
    """Expand the expected code string in set of codes.

    200-204 -> 200, 201, 202, 204
    200, 203 -> 200, 203
    """
    retval = set()
    codes = re.split(', *', codes)
    for code in codes:
        if not code:
            continue
        if '-' in code:
            low, hi = code.split('-')[:2]
            retval.update(
                str(i) for i in range(int(low), int(hi) + 1))
        else:
            retval.add(code)
    return retval


class exception_logger:
    """Wrap a function and log raised exception

    :param logger: the logger to log the exception default is LOG.exception

    :returns: origin value if no exception raised; re-raise the exception if
              any occurred

    """
    def __init__(self, logger=None):
        self.logger = logger

    def __call__(self, func):
        if self.logger is None:
            _LOG = logging.getLogger(func.__module__)
            self.logger = _LOG.exception

        def call(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                with excutils.save_and_reraise_exception():
                    self.logger(e)
            return None
        return call


def map_protocol_to_nftable_protocol(rule_dict):
    rule_dict[constants.PROTOCOL] = (
        constants.L4_PROTOCOL_MAP[rule_dict[constants.PROTOCOL]])
    return rule_dict
